import asyncio
import os
import logging
import random
from stent import Stent, Result

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s', datefmt='%H:%M:%S')
logger = logging.getLogger("BatchExample")

# --- Activities ---

@Stent.durable()
async def download_image(image_id: int) -> str:
    # Simulate network IO
    delay = random.uniform(0.1, 0.5)
    await asyncio.sleep(delay)
    # Simulate occasional network failure
    if random.random() < 0.1:
        raise ConnectionError(f"Network error downloading image {image_id}")
    
    path = f"/tmp/img_{image_id}.jpg"
    logger.info(f"Downloaded image {image_id} to {path} ({delay:.2f}s)")
    return path

@Stent.durable()
async def process_image(path: str) -> str:
    # Simulate CPU intensive work
    await asyncio.sleep(0.2)
    processed_path = path.replace(".jpg", "_bw.jpg")
    logger.info(f"Processed {path} -> {processed_path}")
    return processed_path

@Stent.durable()
async def create_gallery(image_paths: list[str]) -> str:
    await asyncio.sleep(0.5)
    logger.info(f"Creating gallery from {len(image_paths)} images")
    return f"http://example.com/gallery/{len(image_paths)}_images"

# --- Orchestrator ---

@Stent.durable()
async def batch_image_workflow(image_ids: list[int]) -> Result[str, Exception]:
    logger.info(f"Starting batch workflow for {len(image_ids)} images")
    
    # Fan-out: Download all images in parallel
    download_tasks = []
    for img_id in image_ids:
        # We can fire off tasks. If one fails, we might want to handle it.
        # asyncio.gather will raise the first exception by default.
        # In a real batch, maybe we want return_exceptions=True to process partials.
        download_tasks.append(download_image(img_id))
    
    # Wait for downloads
    # To handle partial failures, we would wrap download_image in a safe version or use return_exceptions.
    # Here we assume we want all or nothing for simplicity, but let's use return_exceptions=True to show robustness.
    download_results = await asyncio.gather(*download_tasks, return_exceptions=True)
    
    successful_downloads = []
    for i, res in enumerate(download_results):
        if isinstance(res, Exception):
            logger.warning(f"Failed to download image {image_ids[i]}: {res}")
        else:
            successful_downloads.append(res)
            
    if not successful_downloads:
        return Result.Error(Exception("No images downloaded successfully"))

    # Fan-out: Process downloaded images
    process_tasks = [process_image(path) for path in successful_downloads]
    processed_paths = await asyncio.gather(*process_tasks)
    
    # Fan-in: Create gallery
    gallery_url = await create_gallery(processed_paths)
    
    return Result.Ok(gallery_url)

# --- Runner ---

async def main():
    db_path = "batch_example.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
    
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    
    executor = Stent(backend=backend)
    
    # Start worker with concurrency
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.1, max_concurrency=20))
    
    logger.info("Dispatching workflow...")
    # Process 10 images
    ids = list(range(1, 11))
    exec_id = await executor.dispatch(batch_image_workflow, ids)
    
    # Monitor
    while True:
        state = await executor.state_of(exec_id)
        if state.state in ("completed", "failed", "timed_out"):
            break
        await asyncio.sleep(0.5)
        
    result = await executor.result_of(exec_id)
    logger.info(f"Final Result: {result}")
    
    worker_task.cancel()
    try:
        await worker_task
    except asyncio.CancelledError:
        pass

if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import os
import logging
from stent import Stent, Result, RetryPolicy

logging.basicConfig(level=logging.INFO)

@Stent.durable(retry_policy=RetryPolicy(max_attempts=2, initial_delay=0.1))
async def fail_always():
    print("Executing fail_always")
    raise ValueError("Something went wrong")

@Stent.durable()
async def failing_flow():
    print("Executing failing_flow")
    await fail_always()

async def run():
    db_path = "stent_fail.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    
    executor = Stent(backend=backend)
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.5))
    
    execution_id = await executor.dispatch(failing_flow)
    print(f"Dispatched: {execution_id}")
    
    while True:
        state = await executor.state_of(execution_id)
        if state.state in ("completed", "failed"):
            print(f"Final State: {state.state}")
            break
        await asyncio.sleep(0.5)
        
    result = await executor.result_of(execution_id)
    print(f"Result: {result}")
    
    worker_task.cancel()
    try:
        await worker_task
    except asyncio.CancelledError:
        pass

if __name__ == "__main__":
    asyncio.run(run())



from __future__ import annotations

import base64
import json
from datetime import datetime
from typing import Any, Callable, Sequence

from stent.core import DeadLetterRecord, ExecutionProgress, ExecutionRecord, SignalRecord, TaskRecord, RetryPolicy

PlaceholderFn = Callable[[int], str]


def _encode_bytes(value: bytes | None) -> str | None:
    if value is None:
        return None
    return base64.b64encode(value).decode("ascii")


def _decode_bytes(value: str | None) -> bytes | None:
    if value is None:
        return None
    return base64.b64decode(value.encode("ascii"))


def _datetime_to_str(value: datetime | None) -> str | None:
    if value is None:
        return None
    return value.isoformat()


def _datetime_from_str(value: str | None) -> datetime | None:
    if not value:
        return None
    return datetime.fromisoformat(value)


def _retry_policy_to_dict(policy: RetryPolicy | None) -> dict[str, Any] | None:
    if not policy:
        return None
    return {
        "max_attempts": policy.max_attempts,
        "backoff_factor": policy.backoff_factor,
        "initial_delay": policy.initial_delay,
        "max_delay": policy.max_delay,
        "jitter": policy.jitter,
        "retry_for": [exc.__name__ for exc in policy.retry_for],
    }


def _retry_policy_from_dict(data: dict[str, Any] | None) -> RetryPolicy | None:
    if not data:
        return None
    
    # Import here to avoid circular imports
    from stent.utils.serialization import get_exception_class
    
    # Deserialize retry_for exception classes
    retry_for_names = data.get("retry_for", ["Exception"])
    retry_for_classes: list[type[BaseException]] = []
    for name in retry_for_names:
        exc_class = get_exception_class(name)
        if exc_class not in retry_for_classes:
            retry_for_classes.append(exc_class)
    
    if not retry_for_classes:
        retry_for_classes = [Exception]
    
    return RetryPolicy(
        max_attempts=data.get("max_attempts", 3),
        backoff_factor=data.get("backoff_factor", 2.0),
        initial_delay=data.get("initial_delay", 1.0),
        max_delay=data.get("max_delay", 60.0),
        jitter=data.get("jitter", 0.1),
        retry_for=tuple(retry_for_classes),
    )


def task_record_to_json(task: TaskRecord) -> str:
    payload = {
        "id": task.id,
        "execution_id": task.execution_id,
        "step_name": task.step_name,
        "kind": task.kind,
        "parent_task_id": task.parent_task_id,
        "state": task.state,
        "args": _encode_bytes(task.args),
        "kwargs": _encode_bytes(task.kwargs),
        "result": _encode_bytes(task.result),
        "error": _encode_bytes(task.error),
        "retries": task.retries,
        "created_at": _datetime_to_str(task.created_at),
        "started_at": _datetime_to_str(task.started_at),
        "completed_at": _datetime_to_str(task.completed_at),
        "worker_id": task.worker_id,
        "lease_expires_at": _datetime_to_str(task.lease_expires_at),
        "tags": task.tags,
        "priority": task.priority,
        "queue": task.queue,
        "retry_policy": _retry_policy_to_dict(task.retry_policy),
        "idempotency_key": task.idempotency_key,
        "scheduled_for": _datetime_to_str(task.scheduled_for),
    }
    return json.dumps(payload)


def task_record_from_json(payload: str) -> TaskRecord:
    data = json.loads(payload)
    return TaskRecord(
        id=data["id"],
        execution_id=data["execution_id"],
        step_name=data["step_name"],
        kind=data["kind"],
        parent_task_id=data.get("parent_task_id"),
        state=data["state"],
        args=_decode_bytes(data.get("args")) or b"",
        kwargs=_decode_bytes(data.get("kwargs")) or b"",
        retries=data["retries"],
        created_at=_datetime_from_str(data.get("created_at")) or datetime.now(),
        tags=list(data.get("tags") or []),
        priority=data.get("priority", 0),
        queue=data.get("queue"),
        retry_policy=_retry_policy_from_dict(data.get("retry_policy")),
        result=_decode_bytes(data.get("result")),
        error=_decode_bytes(data.get("error")),
        started_at=_datetime_from_str(data.get("started_at")),
        completed_at=_datetime_from_str(data.get("completed_at")),
        worker_id=data.get("worker_id"),
        lease_expires_at=_datetime_from_str(data.get("lease_expires_at")),
        idempotency_key=data.get("idempotency_key"),
        scheduled_for=_datetime_from_str(data.get("scheduled_for")),
    )


def retry_policy_to_json(policy: RetryPolicy | None) -> str:
    if not policy:
        return "{}"
    return json.dumps(
        {
            "max_attempts": policy.max_attempts,
            "backoff_factor": policy.backoff_factor,
            "initial_delay": policy.initial_delay,
            "max_delay": policy.max_delay,
            "jitter": policy.jitter,
        }
    )


def retry_policy_from_json(value: str | None) -> RetryPolicy:
    data = json.loads(value) if value else {}
    return RetryPolicy(
        max_attempts=data.get("max_attempts", 3),
        backoff_factor=data.get("backoff_factor", 2.0),
        initial_delay=data.get("initial_delay", 1.0),
        max_delay=data.get("max_delay", 60.0),
        jitter=data.get("jitter", 0.1),
    )


def execution_row_values(record: ExecutionRecord) -> tuple[Any, ...]:
    return (
        record.id,
        record.root_function,
        record.state,
        record.args,
        record.kwargs,
        record.result,
        record.error,
        record.retries,
        record.created_at,
        record.started_at,
        record.completed_at,
        record.expiry_at,
        json.dumps(record.tags),
        record.priority,
        record.queue,
    )


def task_row_values(task: TaskRecord) -> tuple[Any, ...]:
    return (
        task.id,
        task.execution_id,
        task.step_name,
        task.kind,
        task.parent_task_id,
        task.state,
        task.args,
        task.kwargs,
        task.result,
        task.error,
        task.retries,
        task.created_at,
        task.started_at,
        task.completed_at,
        task.worker_id,
        task.lease_expires_at,
        json.dumps(task.tags),
        task.priority,
        task.queue,
        task.idempotency_key,
        retry_policy_to_json(task.retry_policy),
        task.scheduled_for,
    )


def _as_datetime(value: Any) -> datetime | None:
    if value is None:
        return None
    if isinstance(value, str):
        return datetime.fromisoformat(value)
    return value


def _loads_tags(value: Any) -> list[str]:
    if isinstance(value, str):
        return list(json.loads(value))
    return list(value or [])


def row_to_progress(row: Any) -> ExecutionProgress:
    return ExecutionProgress(
        step=row["step"],
        status=row["status"],
        started_at=_as_datetime(row["started_at"]),
        completed_at=_as_datetime(row["completed_at"]),
        detail=row["detail"],
    )


def row_to_execution(row: Any, progress: list[ExecutionProgress]) -> ExecutionRecord:
    return ExecutionRecord(
        id=row["id"],
        root_function=row["root_function"],
        state=row["state"],
        args=row["args"],
        kwargs=row["kwargs"],
        result=row["result"],
        error=row["error"],
        retries=row["retries"],
        created_at=_as_datetime(row["created_at"]) or datetime.now(),
        started_at=_as_datetime(row["started_at"]),
        completed_at=_as_datetime(row["completed_at"]),
        expiry_at=_as_datetime(row["expiry_at"]),
        progress=progress,
        tags=_loads_tags(row["tags"]),
        priority=row["priority"],
        queue=row["queue"],
    )


def row_to_task(row: Any) -> TaskRecord:
    return TaskRecord(
        id=row["id"],
        execution_id=row["execution_id"],
        step_name=row["step_name"],
        kind=row["kind"],
        parent_task_id=row["parent_task_id"],
        state=row["state"],
        args=row["args"],
        kwargs=row["kwargs"],
        result=row["result"],
        error=row["error"],
        retries=row["retries"],
        created_at=_as_datetime(row["created_at"]) or datetime.now(),
        started_at=_as_datetime(row["started_at"]),
        completed_at=_as_datetime(row["completed_at"]),
        worker_id=row["worker_id"],
        lease_expires_at=_as_datetime(row["lease_expires_at"]),
        tags=_loads_tags(row["tags"]),
        priority=row["priority"],
        queue=row["queue"],
        idempotency_key=row["idempotency_key"],
        retry_policy=retry_policy_from_json(row["retry_policy"]),
        scheduled_for=_as_datetime(row["scheduled_for"]),
    )


def row_to_signal(row: Any) -> SignalRecord:
    return SignalRecord(
        execution_id=row["execution_id"],
        name=row["name"],
        payload=row["payload"],
        created_at=_as_datetime(row["created_at"]) or datetime.now(),
        consumed=bool(row["consumed"]),
        consumed_at=_as_datetime(row["consumed_at"]),
    )


def row_to_dead_letter(row: Any) -> DeadLetterRecord:
    return DeadLetterRecord(
        task=task_record_from_json(row["data"]),
        reason=row["reason"],
        moved_at=_as_datetime(row["moved_at"]) or datetime.now(),
    )


def qmark_placeholder(_: int) -> str:
    return "?"


def dollar_placeholder(index: int) -> str:
    return f"${index}"


def build_filtered_query(
    *,
    base_select: str,
    filters: Sequence[tuple[str, Any | None]],
    placeholder: PlaceholderFn,
) -> tuple[str, list[Any]]:
    params: list[Any] = []
    conditions: list[str] = []
    for column, value in filters:
        if value is None:
            continue
        params.append(value)
        conditions.append(f"{column} = {placeholder(len(params))}")

    query = base_select
    if conditions:
        query += " WHERE " + " AND ".join(conditions)
    return query, params


def build_filtered_count_query(
    *,
    table: str,
    filters: Sequence[tuple[str, Any | None]],
    placeholder: PlaceholderFn,
) -> tuple[str, list[Any]]:
    return build_filtered_query(
        base_select=f"SELECT COUNT(*) FROM {table}",
        filters=filters,
        placeholder=placeholder,
    )


def build_filtered_list_query(
    *,
    table: str,
    filters: Sequence[tuple[str, Any | None]],
    order_by: str,
    limit: int,
    offset: int,
    placeholder: PlaceholderFn,
) -> tuple[str, list[Any]]:
    query, params = build_filtered_query(
        base_select=f"SELECT * FROM {table}",
        filters=filters,
        placeholder=placeholder,
    )
    params.append(limit)
    limit_token = placeholder(len(params))
    params.append(offset)
    offset_token = placeholder(len(params))
    query += f" ORDER BY {order_by} LIMIT {limit_token} OFFSET {offset_token}"
    return query, params



from typing import Protocol, AsyncIterator, Any, Dict, Optional

class NotificationBackend(Protocol):
    async def notify_task_completed(self, task_id: str) -> None: ...
    async def notify_task_updated(self, task_id: str, state: str) -> None: ...
    def subscribe_to_task(
        self,
        task_id: str,
        *,
        expiry: float | None = None,
    ) -> AsyncIterator[Dict[str, Any]]: ...

    async def notify_execution_updated(self, execution_id: str, state: str) -> None: ...
    def subscribe_to_execution(
        self,
        execution_id: str,
        *,
        expiry: float | None = None,
    ) -> AsyncIterator[Dict[str, Any]]: ...


from dataclasses import dataclass
from typing import Callable, Awaitable, Any, List, ItemsView
from stent.core import RetryPolicy

@dataclass
class FunctionMetadata:
    name: str
    fn: Callable[..., Awaitable[Any]]
    cached: bool
    retry_policy: RetryPolicy
    tags: List[str]
    priority: int
    queue: str | None
    version: str | None
    idempotent: bool
    idempotency_key_func: Callable[..., str] | None
    max_concurrent: int | None = None

class FunctionRegistry:
    def __init__(self):
        self._registry: dict[str, FunctionMetadata] = {}

    def register(self, meta: FunctionMetadata) -> None:
        self._registry[meta.name] = meta

    def get(self, name: str) -> FunctionMetadata | None:
        return self._registry.get(name)

    def items(self) -> ItemsView[str, FunctionMetadata]:
        return self._registry.items()

    def copy(self) -> "FunctionRegistry":
        clone = FunctionRegistry()
        clone._registry = dict(self._registry)
        return clone

    def name_for_function(self, fn: Callable[..., Any]) -> str:
        # Simplistic implementation. In a real app we might need stable names regardless of import path.
        return f"{fn.__module__}:{fn.__qualname__}"

registry = FunctionRegistry()



import unittest
import asyncio
import os
import logging
from datetime import datetime, timedelta
from typing import Any, cast
from stent import Stent, Result
from stent.core import TaskRecord, RetryPolicy, ExecutionRecord
import stent.executor as executor_module
from stent.executor import ExpiryError, _original_sleep
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

logger = logging.getLogger(__name__)

@Stent.durable()
async def quick_task():
    return "quick"

@Stent.durable()
async def slow_task(duration: float):
    await _original_sleep(duration)  # real sleep to simulate slow work
    return "slow"

class TestWaitFor(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"waitfor_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_wait_success(self):
        exec_id = await self.executor.dispatch(quick_task)
        
        # Should block until done
        result = await self.executor.wait_for(exec_id, expiry=10.0) # Increased expiry
        
        self.assertTrue(result.ok)
        self.assertEqual(result.value, "quick")

    async def test_wait_for_timeout(self):
        # We need a longer task than the wait expiry
        @Stent.durable()
        async def slow_task_impl():
            await _original_sleep(5.0)  # real sleep to simulate slow work
            return "slow"
            
        exec_id = await self.executor.dispatch(slow_task_impl)
        
        with self.assertRaises(ExpiryError):
            await self.executor.wait_for(exec_id, expiry=0.5) # Allow 0.5s for dispatch overhead

        # Clean up by waiting properly
        await self.executor.wait_for(exec_id, expiry=10.0)

    async def test_wait_for_already_completed(self):
        exec_id = await self.executor.dispatch(quick_task)
        # Wait manually first
        await self.executor.wait_for(exec_id, expiry=5.0)
    
        # Now call wait_for, should return immediately
        start = asyncio.get_running_loop().time()
        result = await self.executor.wait_for(exec_id, expiry=5.0)
        end = asyncio.get_running_loop().time()
    
        self.assertTrue(result.ok)
        self.assertLess(end - start, 1.0) # Should be instant-ish, but allow 1.0 for remote DB roundtrip



class _PollingDummyBackend:
    def __init__(self):
        self.calls = 0

    async def get_task(self, task_id: str):
        self.calls += 1
        if self.calls >= 4:
            return TaskRecord(
                id=task_id,
                execution_id="exec",
                step_name="test",
                kind="activity",
                parent_task_id=None,
                state="completed",
                args=b"",
                kwargs=b"",
                retries=0,
                created_at=datetime.now(),
                tags=[],
                priority=0,
                queue=None,
                retry_policy=RetryPolicy(),
                result=b"done",
            )
        return None


class TestAdaptivePolling(unittest.IsolatedAsyncioTestCase):
    async def test_waiter_backoff_without_notifications(self):
        backend = _PollingDummyBackend()
        executor = Stent(
            backend=cast(Any, backend),
            poll_min_interval=0.01,
            poll_max_interval=0.04,
            poll_backoff_factor=2.0,
        )
        executor.notification_backend = None

        recorded = []
        original_sleep = executor_module._original_sleep

        async def fake_sleep(delay, result=None):
            recorded.append(delay)
            await asyncio.sleep(0)

        executor_module._original_sleep = cast(Any, fake_sleep)
        try:
            task = await executor._wait_for_task_internal("poll-task")
            self.assertEqual(task.id, "poll-task")
        finally:
            executor_module._original_sleep = original_sleep

        self.assertEqual(recorded, [0.01, 0.02, 0.04])


class _NotificationTimeoutBackend:
    async def notify_task_completed(self, task_id: str):
        return None

    async def notify_task_updated(self, task_id: str, state: str):
        return None

    async def subscribe_to_task(self, task_id: str, *, expiry: float | None = None):
        if expiry is not None:
            try:
                async with asyncio.timeout(expiry):
                    await asyncio.sleep(expiry + 0.05)
            except asyncio.TimeoutError:
                pass
        if False:
            yield {"task_id": task_id, "state": "pending"}

    async def notify_execution_updated(self, execution_id: str, state: str):
        return None

    async def subscribe_to_execution(self, execution_id: str, *, expiry: float | None = None):
        if expiry is not None:
            try:
                async with asyncio.timeout(expiry):
                    await asyncio.sleep(expiry + 0.05)
            except asyncio.TimeoutError:
                pass
        if False:
            yield {"execution_id": execution_id, "state": "running"}


class _AlwaysPendingBackend:
    async def get_task(self, task_id: str):
        return TaskRecord(
            id=task_id,
            execution_id="exec-timeout",
            step_name="test",
            kind="activity",
            parent_task_id=None,
            state="pending",
            args=b"",
            kwargs=b"",
            retries=0,
            created_at=datetime.now(),
            tags=[],
            priority=0,
            queue=None,
            retry_policy=RetryPolicy(),
        )

    async def get_execution(self, execution_id: str):
        return ExecutionRecord(
            id=execution_id,
            root_function="tests.timeout",
            state="running",
            args=b"{}",
            kwargs=b"{}",
            retries=0,
            created_at=datetime.now(),
            started_at=datetime.now(),
            completed_at=None,
            expiry_at=None,
            progress=[],
            tags=[],
            priority=0,
            queue=None,
        )


class _CancelledExecutionBackend:
    async def get_execution(self, execution_id: str):
        return ExecutionRecord(
            id=execution_id,
            root_function="tests.cancelled",
            state="cancelled",
            args=b"{}",
            kwargs=b"{}",
            retries=0,
            created_at=datetime.now(),
            started_at=datetime.now() - timedelta(seconds=1),
            completed_at=datetime.now(),
            expiry_at=None,
            progress=[],
            tags=[],
            priority=0,
            queue=None,
            result=None,
            error=None,
        )


class TestNotificationTimeoutAndCancelled(unittest.IsolatedAsyncioTestCase):
    async def test_wait_for_task_timeout_with_notifications_raises_expiry(self):
        executor = Stent(
            backend=cast(Any, _AlwaysPendingBackend()),
            notification_backend=cast(Any, _NotificationTimeoutBackend()),
        )

        with self.assertRaises(ExpiryError):
            await executor._wait_for_task_internal("task-timeout", expiry=0.05)

    async def test_wait_for_timeout_with_notifications_raises_expiry(self):
        executor = Stent(
            backend=cast(Any, _AlwaysPendingBackend()),
            notification_backend=cast(Any, _NotificationTimeoutBackend()),
        )

        with self.assertRaises(ExpiryError):
            await executor.wait_for("exec-timeout", expiry=0.05)

    async def test_cancelled_result_of_and_wait_for_are_consistent(self):
        executor = Stent(backend=cast(Any, _CancelledExecutionBackend()))

        result_direct = await executor.result_of("exec-cancelled")
        result_wait = await executor.wait_for("exec-cancelled", expiry=1.0)

        self.assertFalse(result_direct.ok)
        self.assertFalse(result_wait.ok)
        self.assertEqual(str(result_direct.error), "Execution cancelled")
        self.assertEqual(str(result_wait.error), "Execution cancelled")



import aiosqlite
import asyncio
import sqlite3
import json
import logging
from datetime import datetime, timedelta
from typing import List, Optional, Any
from stent.backend.base import Backend
from stent.core import ExecutionRecord, TaskRecord, ExecutionProgress, RetryPolicy, SignalRecord, DeadLetterRecord
from stent.backend.utils import (
    build_filtered_count_query,
    build_filtered_list_query,
    execution_row_values,
    qmark_placeholder,
    row_to_dead_letter,
    row_to_execution,
    row_to_progress,
    row_to_signal,
    row_to_task,
    retry_policy_from_json,
    retry_policy_to_json,
    task_record_to_json,
    task_row_values,
)

logger = logging.getLogger(__name__)

def _adapt_datetime(dt: datetime) -> str:
    return dt.isoformat()

def _convert_datetime(val: bytes) -> datetime:
    return datetime.fromisoformat(val.decode("utf-8"))

sqlite3.register_adapter(datetime, _adapt_datetime)
sqlite3.register_converter("datetime", _convert_datetime)
sqlite3.register_converter("TIMESTAMP", _convert_datetime)


class SQLiteBackend(Backend):
    """
    SQLite backend with persistent connection pooling.
    
    Uses a single persistent connection with serialized access via asyncio.Lock
    to respect SQLite's single-writer model while avoiding the overhead of
    opening a new connection for every operation.
    """
    
    def __init__(self, db_path: str, pool_size: int = 1):
        """
        Initialize SQLite backend.
        
        Args:
            db_path: Path to SQLite database file
            pool_size: Number of connections (default 1, SQLite works best with single writer)
        """
        self.db_path = db_path
        self._pool_size = pool_size
        self._connection: Optional[aiosqlite.Connection] = None
        self._lock = asyncio.Lock()
        self._closed = False

    async def _get_connection(self) -> aiosqlite.Connection:
        """Get or create the persistent connection."""
        if self._closed:
            raise RuntimeError("SQLiteBackend has been closed")
        if self._connection is None:
            self._connection = await aiosqlite.connect(
                self.db_path,
                detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
                isolation_level=None,  # Enable manual transaction control
            )
            self._connection.row_factory = aiosqlite.Row
            # Enable WAL mode for better concurrent read performance
            await self._connection.execute("PRAGMA journal_mode=WAL")
            await self._connection.execute("PRAGMA busy_timeout=5000")
        else:
            # Ensure no stale transaction is left open from a cancelled operation
            if self._connection._conn.in_transaction:
                await self._connection.execute("ROLLBACK")
        return self._connection

    async def close(self) -> None:
        """Close the persistent connection and release resources."""
        async with self._lock:
            self._closed = True
            if self._connection is not None:
                await self._connection.close()
                self._connection = None
                logger.info("SQLite connection closed")

    async def init_db(self) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS signals (
                        execution_id TEXT,
                        name TEXT,
                        payload BLOB,
                        created_at TIMESTAMP,
                        consumed BOOLEAN,
                        consumed_at TIMESTAMP,
                        PRIMARY KEY (execution_id, name)
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS executions (
                        id TEXT PRIMARY KEY,
                        root_function TEXT,
                        state TEXT,
                        args BLOB,
                        kwargs BLOB,
                        result BLOB,
                        error BLOB,
                        retries INTEGER,
                        created_at TIMESTAMP,
                        started_at TIMESTAMP,
                        completed_at TIMESTAMP,
                        expiry_at TIMESTAMP,
                        tags TEXT,
                        priority INTEGER,
                        queue TEXT
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS execution_progress (
                        execution_id TEXT,
                        step TEXT,
                        status TEXT,
                        started_at TIMESTAMP,
                        completed_at TIMESTAMP,
                        detail TEXT,
                        ordinal INTEGER PRIMARY KEY AUTOINCREMENT
                    )
                """)
                await db.execute("CREATE INDEX IF NOT EXISTS idx_progress_exec ON execution_progress(execution_id)")
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS tasks (
                        id TEXT PRIMARY KEY,
                        execution_id TEXT,
                        step_name TEXT,
                        kind TEXT,
                        parent_task_id TEXT,
                        state TEXT,
                        args BLOB,
                        kwargs BLOB,
                        result BLOB,
                        error BLOB,
                        retries INTEGER,
                        created_at TIMESTAMP,
                        started_at TIMESTAMP,
                        completed_at TIMESTAMP,
                        worker_id TEXT,
                        lease_expires_at TIMESTAMP,
                        tags TEXT,
                        priority INTEGER,
                        queue TEXT,
                        idempotency_key TEXT,
                        retry_policy TEXT,
                        scheduled_for TIMESTAMP
                    )
                """)
                await db.execute("CREATE INDEX IF NOT EXISTS idx_tasks_state_queue_scheduled ON tasks(state, queue, scheduled_for)")
                await db.execute("CREATE INDEX IF NOT EXISTS idx_tasks_priority_created ON tasks(priority, created_at)")
                await db.execute("CREATE INDEX IF NOT EXISTS idx_tasks_execution ON tasks(execution_id)")
                await db.execute("CREATE INDEX IF NOT EXISTS idx_tasks_step_lease ON tasks(step_name, state, lease_expires_at)")
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS dead_tasks (
                        id TEXT PRIMARY KEY,
                        reason TEXT,
                        moved_at TIMESTAMP,
                        data TEXT -- full JSON dump of task
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS cache (
                        key TEXT PRIMARY KEY,
                        value BLOB,
                        expires_at TIMESTAMP
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS idempotency (
                        key TEXT PRIMARY KEY,
                        value BLOB
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS execution_counters (
                        execution_id TEXT,
                        name TEXT,
                        value REAL NOT NULL,
                        PRIMARY KEY (execution_id, name)
                    )
                """)
                await db.execute("""
                    CREATE TABLE IF NOT EXISTS execution_state (
                        execution_id TEXT,
                        key TEXT,
                        value BLOB,
                        PRIMARY KEY (execution_id, key)
                    )
                """)
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    def _execution_row_values(self, record: ExecutionRecord) -> tuple[Any, ...]:
        return execution_row_values(record)

    async def _insert_execution(self, db: aiosqlite.Connection, record: ExecutionRecord) -> None:
        await db.execute(
            "INSERT INTO executions VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            self._execution_row_values(record),
        )
        for p in record.progress:
            await db.execute(
                "INSERT INTO execution_progress (execution_id, step, status, started_at, completed_at, detail) VALUES (?, ?, ?, ?, ?, ?)",
                (record.id, p.step, p.status, p.started_at, p.completed_at, p.detail),
            )

    def _task_row_values(self, task: TaskRecord) -> tuple[Any, ...]:
        return task_row_values(task)

    async def _insert_task(self, db: aiosqlite.Connection, task: TaskRecord) -> None:
        await db.execute(
            "INSERT INTO tasks VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            self._task_row_values(task),
        )

    async def create_execution(self, record: ExecutionRecord) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await self._insert_execution(db, record)
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def create_execution_with_root_task(self, record: ExecutionRecord, task: TaskRecord) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN IMMEDIATE")
            try:
                await self._insert_execution(db, record)
                await self._insert_task(db, task)
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_execution(self, execution_id: str) -> ExecutionRecord | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT * FROM executions WHERE id = ?", (execution_id,)) as cursor:
                row = await cursor.fetchone()
                if not row:
                    return None
                
                # Fetch progress
                progress = []
                async with db.execute("SELECT * FROM execution_progress WHERE execution_id = ? ORDER BY ordinal", (execution_id,)) as p_cursor:
                    p_rows = await p_cursor.fetchall()
                    for pr in p_rows:
                         progress.append(self._row_to_progress(pr))

                return self._row_to_execution(row, progress)

    async def update_execution(self, record: ExecutionRecord) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute("""
                    UPDATE executions SET
                        state=?, args=?, kwargs=?, result=?, error=?, retries=?,
                        started_at=?, completed_at=?, expiry_at=?, tags=?,
                        priority=?, queue=?
                    WHERE id=?
                """, (
                    record.state, record.args, record.kwargs, record.result, record.error,
                    record.retries, record.started_at, record.completed_at, record.expiry_at,
                    json.dumps(record.tags), record.priority, record.queue, record.id
                ))
                # Do NOT update progress here as it is managed via execution_progress table
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def list_executions(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[ExecutionRecord]:
        async with self._lock:
            db = await self._get_connection()
            query, params = build_filtered_list_query(
                table="executions",
                filters=[("state", state)],
                order_by="created_at DESC",
                limit=limit,
                offset=offset,
                placeholder=qmark_placeholder,
            )
            
            async with db.execute(query, tuple(params)) as cursor:
                rows = await cursor.fetchall()
                results = []
                for row in rows:
                    # For listing, we might skip fetching progress to keep it light
                    results.append(self._row_to_execution(row, progress=[]))
                return results

    async def count_executions(self, state: str | None = None) -> int:
        async with self._lock:
            db = await self._get_connection()
            query, params = build_filtered_count_query(
                table="executions",
                filters=[("state", state)],
                placeholder=qmark_placeholder,
            )

            async with db.execute(query, tuple(params)) as cursor:
                row = await cursor.fetchone()
                return row[0] if row else 0

    async def create_task(self, task: TaskRecord) -> None:
        await self.create_tasks([task])

    async def create_tasks(self, tasks: List[TaskRecord]) -> None:
        if not tasks:
            return
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.executemany(
                    "INSERT INTO tasks VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    [self._task_row_values(task) for task in tasks],
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def count_tasks(self, queue: str | None = None, state: str | None = None) -> int:
        async with self._lock:
            db = await self._get_connection()
            query, params = build_filtered_count_query(
                table="tasks",
                filters=[("queue", queue), ("state", state)],
                placeholder=qmark_placeholder,
            )
            
            async with db.execute(query, tuple(params)) as cursor:
                row = await cursor.fetchone()
                return row[0] if row else 0

    async def get_task(self, task_id: str) -> TaskRecord | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT * FROM tasks WHERE id = ?", (task_id,)) as cursor:
                row = await cursor.fetchone()
                if not row:
                    return None
                return self._row_to_task(row)

    async def update_task(self, task: TaskRecord) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute("""
                    UPDATE tasks SET
                        state=?, result=?, error=?, retries=?, started_at=?, completed_at=?,
                        worker_id=?, lease_expires_at=?, scheduled_for=?
                    WHERE id=?
                """, (
                    task.state, task.result, task.error, task.retries, task.started_at,
                    task.completed_at, task.worker_id, task.lease_expires_at, task.scheduled_for, task.id
                ))
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def list_tasks(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[TaskRecord]:
        async with self._lock:
            db = await self._get_connection()
            query, params = build_filtered_list_query(
                table="tasks",
                filters=[("state", state)],
                order_by="created_at DESC",
                limit=limit,
                offset=offset,
                placeholder=qmark_placeholder,
            )
            
            async with db.execute(query, tuple(params)) as cursor:
                rows = await cursor.fetchall()
                return [self._row_to_task(row) for row in rows]

    async def claim_next_task(
        self,
        *,
        worker_id: str,
        queues: List[str] | None = None,
        tags: List[str] | None = None,
        now: datetime | None = None,
        lease_duration: timedelta | None = None,
        concurrency_limits: dict[str, int] | None = None,
    ) -> TaskRecord | None:
        if now is None:
            now = datetime.now()
        if lease_duration is None:
            lease_duration = timedelta(minutes=5)
            
        expires_at = now + lease_duration
        worker_tags_set = set(tags or [])
        should_filter_by_tags = bool(tags)
        
        # Helper for queues condition
        queue_clause = ""
        params: List[Any] = [now, now]
        if queues:
            placeholders = ",".join(["?"] * len(queues))
            queue_clause = f"AND (queue IN ({placeholders}) OR queue IS NULL)"
            params.extend(queues)
        else:
            queue_clause = "AND 1=1"

        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN IMMEDIATE")
            try:
                query = f"""
                    SELECT * FROM tasks
                    WHERE (
                        state='pending'
                        OR (state='running' AND lease_expires_at < ?)
                    )
                    AND (scheduled_for IS NULL OR scheduled_for <= ?)
                    AND kind != 'signal'
                    {queue_clause}
                    ORDER BY priority DESC, created_at ASC
                    LIMIT 50
                """
                async with db.execute(query, tuple(params)) as cursor:
                    candidates = await cursor.fetchall()

                if not candidates:
                    await db.execute("ROLLBACK")
                    return None

                for row in candidates:
                    if should_filter_by_tags:
                        task_tags = json.loads(row["tags"]) if row["tags"] else []
                        if task_tags and worker_tags_set.isdisjoint(task_tags):
                            continue

                    step_name = row["step_name"]
                    limit = concurrency_limits.get(step_name) if concurrency_limits else None

                    if limit is not None:
                        count_query = """
                            SELECT COUNT(*) FROM tasks 
                            WHERE step_name = ? 
                            AND state = 'running' 
                            AND lease_expires_at > ?
                        """
                        async with db.execute(count_query, (step_name, now)) as count_cursor:
                            count_row = await count_cursor.fetchone()
                            current_count = count_row[0] if count_row else 0

                        if current_count >= limit:
                            continue

                    claim_query = """
                        UPDATE tasks
                        SET state='running', worker_id=?, lease_expires_at=?, started_at=?
                        WHERE id = ?
                        AND (
                            state='pending'
                            OR (state='running' AND lease_expires_at < ?)
                        )
                        RETURNING *
                    """
                    claim_params = (worker_id, expires_at, now, row["id"], now)

                    async with db.execute(claim_query, claim_params) as claim_cursor:
                        claimed_row = await claim_cursor.fetchone()
                        if claimed_row:
                            await db.execute("COMMIT")
                            return self._row_to_task(claimed_row)

                await db.execute("ROLLBACK")
                return None
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def renew_task_lease(
        self,
        task_id: str,
        worker_id: str,
        lease_duration: timedelta,
    ) -> bool:
        new_expiry = datetime.now() + lease_duration
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                cursor = await db.execute(
                    """
                    UPDATE tasks
                    SET lease_expires_at=?
                    WHERE id=? AND worker_id=? AND state='running'
                    """,
                    (new_expiry, task_id, worker_id),
                )
                await db.execute("COMMIT")
                return (cursor.rowcount or 0) > 0
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def list_tasks_for_execution(self, execution_id: str) -> List[TaskRecord]:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT * FROM tasks WHERE execution_id = ?", (execution_id,)) as cursor:
                rows = await cursor.fetchall()
                return [self._row_to_task(row) for row in rows]

    async def append_progress(self, execution_id: str, progress: ExecutionProgress) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    "INSERT INTO execution_progress (execution_id, step, status, started_at, completed_at, detail) VALUES (?, ?, ?, ?, ?, ?)",
                    (execution_id, progress.step, progress.status, progress.started_at, progress.completed_at, progress.detail)
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_cached_result(self, cache_key: str) -> bytes | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT value, expires_at FROM cache WHERE key = ?", (cache_key,)) as cursor:
                row = await cursor.fetchone()
                if row:
                    val, expires_at = row[0], row[1]
                    logger.debug(f"Fetched from cache: key={cache_key}, expires_at={expires_at}, value_len={len(val) if val else 0}")
                    if expires_at and datetime.fromisoformat(expires_at) < datetime.now():
                        logger.debug(f"Cache expired for key={cache_key}")
                        return None
                    return val
                logger.debug(f"Cache miss for key={cache_key}")
        return None

    async def set_cached_result(self, cache_key: str, value: bytes, ttl: timedelta | None = None) -> None:
        expires_at = None
        if ttl:
            expires_at = datetime.now() + ttl
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    "INSERT OR REPLACE INTO cache (key, value, expires_at) VALUES (?, ?, ?)",
                    (cache_key, value, expires_at)
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_idempotency_result(self, idempotency_key: str) -> bytes | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT value FROM idempotency WHERE key = ?", (idempotency_key,)) as cursor:
                row = await cursor.fetchone()
                return row[0] if row else None

    async def set_idempotency_result(self, idempotency_key: str, value: bytes) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    "INSERT OR REPLACE INTO idempotency (key, value) VALUES (?, ?)",
                    (idempotency_key, value)
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def move_task_to_dead_letter(self, task: TaskRecord, reason: str) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                payload = task_record_to_json(task)
                await db.execute(
                    "INSERT INTO dead_tasks (id, reason, moved_at, data) VALUES (?, ?, ?, ?)",
                    (task.id, reason, datetime.now(), payload)
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    def _row_to_dead_letter(self, row: Any) -> DeadLetterRecord:
        return row_to_dead_letter(row)

    async def list_dead_tasks(self, limit: int = 50) -> List[DeadLetterRecord]:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute(
                "SELECT * FROM dead_tasks ORDER BY moved_at DESC LIMIT ?",
                (limit,),
            ) as cursor:
                rows = await cursor.fetchall()
                return [self._row_to_dead_letter(row) for row in rows]

    async def count_dead_tasks(self) -> int:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT COUNT(*) FROM dead_tasks") as cursor:
                row = await cursor.fetchone()
                return row[0] if row else 0

    async def get_dead_task(self, task_id: str) -> DeadLetterRecord | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute(
                "SELECT * FROM dead_tasks WHERE id = ?",
                (task_id,),
            ) as cursor:
                row = await cursor.fetchone()
                if not row:
                    return None
                return self._row_to_dead_letter(row)

    async def delete_dead_task(self, task_id: str) -> bool:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                cursor = await db.execute(
                    "DELETE FROM dead_tasks WHERE id = ?",
                    (task_id,),
                )
                await db.execute("COMMIT")
                return (cursor.rowcount or 0) > 0
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def cleanup_executions(self, older_than: datetime) -> int:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                where_clause = "completed_at IS NOT NULL AND completed_at < ? AND state IN ('completed', 'failed', 'timed_out', 'cancelled')"
                
                # Delete dependents using subquery
                await db.execute(f"DELETE FROM tasks WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", (older_than,))
                await db.execute(f"DELETE FROM execution_progress WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", (older_than,))
                await db.execute(f"DELETE FROM signals WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", (older_than,))
                await db.execute(f"DELETE FROM execution_counters WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", (older_than,))
                await db.execute(f"DELETE FROM execution_state WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", (older_than,))
                
                # Delete executions
                cursor = await db.execute(f"DELETE FROM executions WHERE {where_clause}", (older_than,))
                count = cursor.rowcount or 0
                await db.execute("COMMIT")
                return count
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def cleanup_dead_letters(self, older_than: datetime) -> int:
        """Remove dead letter records older than the specified datetime."""
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                cursor = await db.execute(
                    "DELETE FROM dead_tasks WHERE moved_at < ?",
                    (older_than,),
                )
                await db.execute("COMMIT")
                return cursor.rowcount or 0
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def create_signal(self, signal: SignalRecord) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    "INSERT OR REPLACE INTO signals (execution_id, name, payload, created_at, consumed, consumed_at) VALUES (?, ?, ?, ?, ?, ?)",
                    (signal.execution_id, signal.name, signal.payload, signal.created_at, signal.consumed, signal.consumed_at)
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_signal(self, execution_id: str, name: str) -> SignalRecord | None:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute("SELECT * FROM signals WHERE execution_id = ? AND name = ?", (execution_id, name)) as cursor:
                row = await cursor.fetchone()
                if not row:
                    return None
                return row_to_signal(row)

    async def ensure_execution_counters(self, execution_id: str, counters: dict[str, int | float]) -> None:
        if not counters:
            return
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.executemany(
                    """
                    INSERT INTO execution_counters (execution_id, name, value)
                    VALUES (?, ?, ?)
                    ON CONFLICT(execution_id, name) DO NOTHING
                    """,
                    [(execution_id, name, float(value)) for name, value in counters.items()],
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def increment_execution_counter(self, execution_id: str, name: str, amount: int | float) -> float:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    """
                    INSERT INTO execution_counters (execution_id, name, value)
                    VALUES (?, ?, ?)
                    ON CONFLICT(execution_id, name) DO UPDATE SET value = value + excluded.value
                    """,
                    (execution_id, name, float(amount)),
                )
                async with db.execute(
                    "SELECT value FROM execution_counters WHERE execution_id = ? AND name = ?",
                    (execution_id, name),
                ) as cursor:
                    row = await cursor.fetchone()
                await db.execute("COMMIT")
                return float(row[0]) if row else 0.0
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_execution_counters(self, execution_id: str) -> dict[str, int | float]:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute(
                "SELECT name, value FROM execution_counters WHERE execution_id = ?",
                (execution_id,),
            ) as cursor:
                rows = await cursor.fetchall()
                return {row["name"]: row["value"] for row in rows}

    async def ensure_execution_state_values(self, execution_id: str, values: dict[str, bytes]) -> None:
        if not values:
            return
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.executemany(
                    """
                    INSERT INTO execution_state (execution_id, key, value)
                    VALUES (?, ?, ?)
                    ON CONFLICT(execution_id, key) DO NOTHING
                    """,
                    [(execution_id, key, value) for key, value in values.items()],
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def set_execution_state_value(self, execution_id: str, key: str, value: bytes) -> None:
        async with self._lock:
            db = await self._get_connection()
            await db.execute("BEGIN")
            try:
                await db.execute(
                    """
                    INSERT INTO execution_state (execution_id, key, value)
                    VALUES (?, ?, ?)
                    ON CONFLICT(execution_id, key) DO UPDATE SET value = excluded.value
                    """,
                    (execution_id, key, value),
                )
                await db.execute("COMMIT")
            except Exception:
                await db.execute("ROLLBACK")
                raise

    async def get_execution_state_values(self, execution_id: str) -> dict[str, bytes]:
        async with self._lock:
            db = await self._get_connection()
            async with db.execute(
                "SELECT key, value FROM execution_state WHERE execution_id = ?",
                (execution_id,),
            ) as cursor:
                rows = await cursor.fetchall()
                return {row["key"]: row["value"] for row in rows}


    def _progress_to_dict(self, p: ExecutionProgress) -> dict:
        return {
            "step": p.step,
            "status": p.status,
            "started_at": p.started_at.isoformat() if p.started_at else None,
            "completed_at": p.completed_at.isoformat() if p.completed_at else None,
            "detail": p.detail
        }

    def _policy_to_json(self, p: RetryPolicy | None) -> str:
        return retry_policy_to_json(p)

    def _json_to_policy(self, s: str) -> RetryPolicy:
        return retry_policy_from_json(s)

    def _row_to_execution(self, row: Any, progress: List[ExecutionProgress]) -> ExecutionRecord:
        return row_to_execution(row, progress)

    def _row_to_progress(self, row: Any) -> ExecutionProgress:
        return row_to_progress(row)

    def _row_to_task(self, row: Any) -> TaskRecord:
        return row_to_task(row)



from __future__ import annotations

from typing import Protocol


class MetricsRecorder(Protocol):
    """
    Receives callbacks when important Stent events occur so hosts can export
    metrics to Prometheus, StatsD, etc.
    """

    def task_claimed(self, *, queue: str | None, step_name: str, kind: str) -> None: ...

    def task_completed(
        self,
        *,
        queue: str | None,
        step_name: str,
        kind: str,
        duration_s: float,
    ) -> None: ...

    def task_failed(
        self,
        *,
        queue: str | None,
        step_name: str,
        kind: str,
        reason: str,
        retrying: bool,
    ) -> None: ...

    def dead_lettered(
        self,
        *,
        queue: str | None,
        step_name: str,
        kind: str,
        reason: str,
    ) -> None: ...

    def lease_renewed(self, *, task_id: str, success: bool) -> None: ...


class NoOpMetricsRecorder:
    def task_claimed(self, **_kwargs) -> None:
        pass

    def task_completed(self, **_kwargs) -> None:
        pass

    def task_failed(self, **_kwargs) -> None:
        pass

    def dead_lettered(self, **_kwargs) -> None:
        pass

    def lease_renewed(self, **_kwargs) -> None:
        pass



import asyncio
import os
import unittest

from stent import Stent
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend


@Stent.durable()
async def context_bar():
    ctx = Stent.context()
    ctx.counters("progress").add(1)
    ctx.counters("abc").add(40)
    ctx.state("phase").set("bar_done")
    await asyncio.sleep(0)


@Stent.durable()
async def context_foo():
    ctx = Stent.context(
        counters={
            "progress": 0,
            "volume": 30,
        },
        state={
            "owner": "foo",
        },
    )
    ctx.state("status").set("running")
    await context_bar()
    return "ok"


class TestExecutionContext(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"ctx_{os.getpid()}_{id(self)}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.05))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_context_counters_and_custom_state(self):
        exec_id = await self.executor.dispatch(context_foo)
        result = await self.executor.wait_for(exec_id, expiry=5.0)
        self.assertTrue(result.ok)

        state = await self.executor.state_of(exec_id)
        self.assertEqual(state.counters["progress"], 1)
        self.assertEqual(state.counters["volume"], 30)
        self.assertEqual(state.counters["abc"], 40)

        self.assertEqual(state.custom_state["owner"], "foo")
        self.assertEqual(state.custom_state["status"], "running")
        self.assertEqual(state.custom_state["phase"], "bar_done")



from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Generic, TypeVar, Any, Literal, Optional, List
import random

T = TypeVar("T")
E = TypeVar("E")

@dataclass
class Result(Generic[T, E]):
    ok: bool
    value: T | None
    error: E | None

    @classmethod
    def Ok(cls, value: T) -> "Result[T, E]":
        return cls(ok=True, value=value, error=None)

    @classmethod
    def Error(cls, error: E) -> "Result[T, E]":
        return cls(ok=False, value=None, error=error)

    def or_raise(self) -> T:
        if not self.ok:
            if isinstance(self.error, BaseException):
                raise self.error
            raise Exception(self.error)
        return self.value  # type: ignore

@dataclass
class RetryPolicy:
    max_attempts: int = 3
    backoff_factor: float = 2.0
    initial_delay: float = 1.0
    max_delay: float = 60.0
    jitter: float = 0.1
    retry_for: tuple[type[BaseException], ...] = (Exception,)

def compute_retry_delay(policy: RetryPolicy, attempt: int, *, rng: random.Random = random.Random()) -> float:
    # attempt is 1-based (1 = first retry)
    # wait = initial * (factor ^ (attempt - 1))
    delay = policy.initial_delay * (policy.backoff_factor ** (attempt - 1))
    delay = min(delay, policy.max_delay)
    
    if policy.jitter > 0:
        # randomize between delay * (1 - jitter) and delay * (1 + jitter) ?? 
        # Or usually just +/- jitter or 0 to jitter. 
        # Prompt says "fraction of delay to randomize (0-1)".
        # Let's assume standard implementation: delay +/- (delay * jitter)
        jitter_amount = delay * policy.jitter
        delay = delay + rng.uniform(-jitter_amount, jitter_amount)
    
    return max(0.0, delay)

@dataclass
class ExecutionProgress:
    step: str
    status: Literal["dispatched", "running", "completed", "failed", "cache_hit"]
    started_at: datetime | None = None
    completed_at: datetime | None = None
    detail: str | None = None

@dataclass
class ExecutionRecord:
    id: str
    root_function: str
    state: Literal["pending", "running", "completed", "failed", "timed_out", "cancelling", "cancelled"]
    args: bytes
    kwargs: bytes
    retries: int
    created_at: datetime
    started_at: datetime | None
    completed_at: datetime | None
    expiry_at: datetime | None
    progress: List[ExecutionProgress]
    tags: List[str]
    priority: int
    queue: str | None
    result: bytes | None = None
    error: bytes | None = None

@dataclass
class TaskRecord:
    id: str
    execution_id: str
    step_name: str
    kind: Literal["orchestrator", "activity", "signal"]
    parent_task_id: str | None
    state: Literal["pending", "running", "completed", "failed"]
    args: bytes
    kwargs: bytes
    retries: int
    created_at: datetime
    tags: List[str]
    priority: int
    queue: str | None
    retry_policy: RetryPolicy | None
    result: bytes | None = None
    error: bytes | None = None
    started_at: datetime | None = None
    completed_at: datetime | None = None
    worker_id: str | None = None
    lease_expires_at: datetime | None = None
    idempotency_key: str | None = None
    scheduled_for: datetime | None = None

@dataclass
class SignalRecord:
    execution_id: str
    name: str
    payload: bytes
    created_at: datetime
    consumed: bool = False
    consumed_at: datetime | None = None

@dataclass
class DeadLetterRecord:
    task: TaskRecord
    reason: str
    moved_at: datetime

    @property
    def id(self) -> str:
        return self.task.id

    @property
    def execution_id(self) -> str:
        return self.task.execution_id

@dataclass
class ExecutionState:
    id: str
    state: str
    result: Result[Any, Any] | None
    started_at: datetime | None
    completed_at: datetime | None
    retries: int
    progress: List[ExecutionProgress]
    tags: List[str]
    priority: int
    queue: str | None
    counters: dict[str, int | float] = field(default_factory=dict)
    custom_state: dict[str, Any] = field(default_factory=dict)

    @property
    def progress_str(self) -> str:
        # Simple implementation
        parts: List[str] = []
        for p in self.progress:
            if not p.completed_at:
                continue
            duration = ""
            if p.completed_at and p.started_at:
                diff = (p.completed_at - p.started_at).total_seconds()
                duration = f"({diff:.1f}s)"
            elif p.started_at:
                 duration = "(...)"
            parts.append(f"{p.step}{duration}")
        return f"{self.progress[0].step if len(self.progress) else '(begin)'} > " + " > ".join(parts)



import asyncio
import os
import logging
from stent import Stent, Result, RetryPolicy

logging.basicConfig(level=logging.INFO)

# Define functions
@Stent.durable()
async def add(x: int, y: int) -> int:
    print(f"Executing add({x}, {y})")
    await asyncio.sleep(0.5)
    return x + y

@Stent.durable()
async def main_flow(start_val: int) -> Result[int, Exception]:
    print(f"Executing main_flow({start_val})")
    # Calling durable function directly
    r1 = await add(start_val, 10)
    # Chain another
    r2 = await add(r1, 5)
    return Result.Ok(r2)

async def run():
    db_path = "stent.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    # We need to initialize the DB. The interface doesn't expose init_db strictly 
    # but SQLiteBackend has it.
    await backend.init_db()
    
    executor = Stent(backend=backend)
    
    # Start worker
    print("Starting worker...")
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.5))
    
    # Dispatch
    print("Dispatching workflow...")
    execution_id = await executor.dispatch(main_flow, 100)
    print(f"Dispatched execution ID: {execution_id}")
    
    # Wait for result
    while True:
        state = await executor.state_of(execution_id)
        print(f"Execution State: {state.state}")
        print(f"Progress: {state.progress_str}")
        
        if state.state in ("completed", "failed"):
            break
        await asyncio.sleep(1)
        
    result = await executor.result_of(execution_id)
    print(f"Final Result: {result}")
    
    worker_task.cancel()
    try:
        await worker_task
    except asyncio.CancelledError:
        pass

if __name__ == "__main__":
    asyncio.run(run())



import unittest
import json
from datetime import timedelta, timezone
from stent.core import Result, RetryPolicy, compute_retry_delay
from stent.utils.serialization import JsonSerializer
from stent.utils.time import parse_duration, now_utc

class TestCore(unittest.TestCase):
    def test_result_serialization(self):
        serializer = JsonSerializer()
        
        # Test Ok
        r1 = Result.Ok({"foo": "bar"})
        encoded = serializer.dumps(r1)
        decoded = serializer.loads(encoded)
        self.assertTrue(decoded.ok)
        self.assertEqual(decoded.value, {"foo": "bar"})
        self.assertIsNone(decoded.error)

        # Test Error
        r2 = Result.Error("Something bad")
        encoded = serializer.dumps(r2)
        decoded = serializer.loads(encoded)
        self.assertFalse(decoded.ok)
        self.assertEqual(decoded.error, "Something bad")
        
    def test_retry_policy_calculation(self):
        policy = RetryPolicy(
            initial_delay=1.0,
            backoff_factor=2.0,
            jitter=0.0 # disable jitter for deterministic check
        )
        
        # attempt 1: 1.0 * (2^(1-1)) = 1.0
        self.assertEqual(compute_retry_delay(policy, 1), 1.0)
        # attempt 2: 1.0 * (2^(2-1)) = 2.0
        self.assertEqual(compute_retry_delay(policy, 2), 2.0)
        # attempt 3: 1.0 * (2^(3-1)) = 4.0
        self.assertEqual(compute_retry_delay(policy, 3), 4.0)
        
    def test_retry_policy_jitter(self):
        policy = RetryPolicy(
            initial_delay=1.0,
            jitter=0.1
        )
        # Should be between 0.9 and 1.1
        delay = compute_retry_delay(policy, 1)
        self.assertTrue(0.9 <= delay <= 1.1)

    def test_parse_duration(self):
        self.assertEqual(parse_duration("30s"), timedelta(seconds=30))
        self.assertEqual(parse_duration("5m"), timedelta(minutes=5))
        self.assertEqual(parse_duration("1h"), timedelta(hours=1))
        self.assertEqual(parse_duration("1h30m"), timedelta(hours=1, minutes=30))
        self.assertEqual(parse_duration("2d8h"), timedelta(days=2, hours=8))
        self.assertEqual(parse_duration({"minutes": 5}), timedelta(minutes=5))
        self.assertEqual(parse_duration(timedelta(seconds=12)), timedelta(seconds=12))

    def test_parse_duration_rejects_invalid_formats(self):
        invalid_values = [
            "",
            "10x",
            "10sfoo",
            "foo10s",
            "1h 30m",
        ]

        for value in invalid_values:
            with self.assertRaises(ValueError):
                parse_duration(value)

    def test_now_utc_is_timezone_aware_utc(self):
        current = now_utc()
        self.assertIsNotNone(current.tzinfo)
        self.assertEqual(current.tzinfo, timezone.utc)



import unittest
import asyncio
import os
from datetime import datetime, timedelta
import stent # import the package to access stent.sleep
from stent.executor import Stent, Backends
from stent.registry import registry
from stent.utils.time import parse_duration
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

# Define some tasks
@Stent.durable()
async def quick_task(x: int):
    return x * 2

@Stent.durable()
async def sleeping_workflow():
    await stent.sleep("1s")
    return "done"

@Stent.durable()
async def asyncio_sleep_workflow():
    await asyncio.sleep(1.5)  # above threshold → becomes durable
    return "slept"

@Stent.durable()
async def short_asyncio_sleep_workflow():
    await asyncio.sleep(0.05)  # below threshold → stays real sleep
    return "quick"

class TestScheduling(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"scheduling_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.stent = Stent(backend=self.backend)

    async def asyncTearDown(self):
        await self.stent.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_parse_duration_extensions(self):
        self.assertEqual(parse_duration("2d8h"), timedelta(days=2, hours=8))
        self.assertEqual(parse_duration("1m30s"), timedelta(minutes=1, seconds=30))
        self.assertEqual(parse_duration({"minutes": 5}), timedelta(minutes=5))

    async def test_schedule_delayed_execution(self):
        # Schedule to run in 2 seconds
        exec_id = await self.stent.schedule("2s", quick_task, 10)
        
        # Immediately check state
        state = await self.stent.state_of(exec_id)
        self.assertEqual(state.state, "pending")
        
        # Start a worker in background
        worker_task = asyncio.create_task(self.stent.serve(poll_interval=0.1))
        
        try:
            # Wait 1 second - should still be pending (not picked up)
            # We can check DB tasks
            await asyncio.sleep(1)
            # We need to verify it wasn't picked up.
            # We can list tasks.
            tasks = await self.backend.list_tasks_for_execution(exec_id)
            # The orchestrator task should be pending.
            # Depending on timing, if serve loop is fast, it checks "scheduled_for".
            # If scheduled_for > now, it won't pick it up.
            task = tasks[0]
            self.assertEqual(task.state, "pending")
            
            # Wait another 2.5 seconds (total 3.5s) to be safe with remote DB
            await asyncio.sleep(2.5)
            
            # Should be done now
            # Use wait_for to be robust
            result = await self.stent.wait_for(exec_id, expiry=5.0)
            self.assertEqual(result.or_raise(), 20)
            
        finally:
            worker_task.cancel()
            try:
                await worker_task
            except asyncio.CancelledError:
                pass

    async def test_stent_sleep(self):
        # We need to register sleeping_workflow manually if it's not picked up? 
        # Decorator handles it.
        
        start = datetime.now()
        exec_id = await self.stent.dispatch(sleeping_workflow)
        
        worker_task = asyncio.create_task(self.stent.serve(poll_interval=0.1))
        
        try:
            result = await self.stent.wait_for(exec_id, expiry=10.0) # Increased expiry
            duration = (datetime.now() - start).total_seconds()
            
            self.assertEqual(result.or_raise(), "done")
            self.assertGreater(duration, 1.0)
            
        finally:
            worker_task.cancel()
            try:
                await worker_task
            except asyncio.CancelledError:
                pass

    async def test_asyncio_sleep_becomes_durable_above_threshold(self):
        """asyncio.sleep(1.5) inside a durable function should become durable sleep."""
        start = datetime.now()
        exec_id = await self.stent.dispatch(asyncio_sleep_workflow)

        worker_task = asyncio.create_task(self.stent.serve(poll_interval=0.1))
        try:
            result = await self.stent.wait_for(exec_id, expiry=10.0)
            duration = (datetime.now() - start).total_seconds()
            self.assertEqual(result.or_raise(), "slept")
            self.assertGreater(duration, 1.0)

            # Verify a sleep task was created (proof it went through durable path)
            tasks = await self.backend.list_tasks_for_execution(exec_id)
            sleep_tasks = [t for t in tasks if t.step_name == "stent.sleep"]
            self.assertEqual(len(sleep_tasks), 1)
        finally:
            worker_task.cancel()
            try:
                await worker_task
            except asyncio.CancelledError:
                pass

    async def test_asyncio_sleep_below_threshold_stays_real(self):
        """asyncio.sleep(0.05) inside a durable function should stay as real sleep."""
        exec_id = await self.stent.dispatch(short_asyncio_sleep_workflow)

        worker_task = asyncio.create_task(self.stent.serve(poll_interval=0.1))
        try:
            result = await self.stent.wait_for(exec_id, expiry=5.0)
            self.assertEqual(result.or_raise(), "quick")

            # Verify no sleep task was created (stayed as real asyncio.sleep)
            tasks = await self.backend.list_tasks_for_execution(exec_id)
            sleep_tasks = [t for t in tasks if t.step_name == "stent.sleep"]
            self.assertEqual(len(sleep_tasks), 0)
        finally:
            worker_task.cancel()
            try:
                await worker_task
            except asyncio.CancelledError:
                pass


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import asyncio
import functools
import uuid
import logging
from datetime import datetime, timedelta
from typing import Callable, Awaitable, Any, ClassVar, List, Literal, Optional, cast
from contextvars import ContextVar

from stent.core import (
    Result, RetryPolicy, TaskRecord, ExecutionProgress, 
    ExecutionState, DeadLetterRecord
)
from stent.backend.base import Backend
from stent.notifications.base import NotificationBackend
from stent.registry import registry, FunctionMetadata, FunctionRegistry
from stent.utils.serialization import Serializer, JsonSerializer
from stent.utils.idempotency import default_idempotency_key
from stent.utils.time import parse_duration
from stent.executor_signals import persist_signal_and_wake_waiter, resolve_signal_wait
from stent.executor_orchestration import (
    build_dispatch_records,
    execute_map_batch,
    normalize_dispatch_timing,
    persist_dispatch_records,
    schedule_activity_and_wait,
)
from stent.executor_wait import wait_for_execution_terminal, wait_for_task_terminal
from stent.executor_worker import lease_heartbeat_loop, run_claimed_task
from stent.executor_context import ExecutionContext, current_execution_context
from stent.metrics import MetricsRecorder, NoOpMetricsRecorder

from dataclasses import dataclass, field, replace

logger = logging.getLogger(__name__)

class StentLogFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        record.stent_execution_id = current_execution_id.get()
        record.stent_task_id = current_task_id.get()
        record.stent_worker_id = current_worker_id.get()
        return True

def install_structured_logging(target: logging.Logger | None = None) -> None:
    """
    Adds a logging.Filter that injects Stent context (execution/task id) into
    every log record. Integrate it with your formatter via
    %(stent_execution_id)s etc.
    """
    logger_obj = target or logging.getLogger("stent")
    for existing in getattr(logger_obj, "filters", []):
        if isinstance(existing, StentLogFilter):
            return
    logger_obj.addFilter(StentLogFilter())

install_structured_logging(logger)

class ExpiryError(TimeoutError):
    pass


class UnregisteredFunctionError(RuntimeError):
    def __init__(self, function_name: str):
        message = (
            f"No durable registration found for '{function_name}'. "
            "Decorate the function with @Stent.durable() or register it on the "
            "FunctionRegistry provided to the executor."
        )
        super().__init__(message)
        self.function_name = function_name

@dataclass
class PermitHolder:
    sem: asyncio.Semaphore
    held: bool = True

    def release(self):
        if self.held:
            self.sem.release()
            self.held = False
    
    async def acquire(self):
        if not self.held:
            await self.sem.acquire()
            self.held = True

@dataclass(eq=False)
class WorkerLifecycle:
    """
    Represents the lifecycle of a long-running worker loop. Callers can use it
    to coordinate readiness / draining with their process manager (e.g. K8s).
    """
    name: str | None = None
    ready_event: asyncio.Event = field(default_factory=asyncio.Event)
    draining_event: asyncio.Event = field(default_factory=asyncio.Event)
    stopped_event: asyncio.Event = field(default_factory=asyncio.Event)
    stop_event: asyncio.Event = field(default_factory=asyncio.Event)
    state: Literal["starting", "ready", "draining", "stopped"] = "starting"

    def __hash__(self) -> int:
        return id(self)

    def reset(self) -> None:
        self.ready_event.clear()
        self.draining_event.clear()
        self.stopped_event.clear()
        self.stop_event.clear()
        self.state = "starting"

    def mark_ready(self) -> None:
        self.state = "ready"
        self.ready_event.set()

    def mark_draining(self) -> None:
        if self.state != "draining":
            self.state = "draining"
            self.draining_event.set()

    def mark_stopped(self) -> None:
        self.state = "stopped"
        self.stopped_event.set()

    def request_drain(self) -> None:
        self.stop_event.set()

    async def wait_until_ready(self) -> None:
        await self.ready_event.wait()

    async def wait_until_stopped(self) -> None:
        await self.stopped_event.wait()

# Backends helper
class Backends:
    @staticmethod
    def SQLiteBackend(path: str) -> Backend:
        from stent.backend.sqlite import SQLiteBackend
        return SQLiteBackend(path)

    @staticmethod
    def MongoBackend(url: str, db_name: str) -> Backend:
        # Placeholder
        raise NotImplementedError("Mongo backend not implemented yet")

    @staticmethod
    def PostgresBackend(dsn: str) -> Backend:
        from stent.backend.postgres import PostgresBackend
        return PostgresBackend(dsn)

class Notifications:
    @staticmethod
    def RedisBackend(url: str) -> NotificationBackend:
        from stent.notifications.redis import RedisBackend
        return RedisBackend(url)

# Capture original sleep before any patching
_original_sleep = asyncio.sleep

class DurableFunction:
    """Wrapper returned by @Stent.durable(). Callable like a normal async
    function, but also exposes .map() and .starmap() for batch dispatch."""

    def __init__(self, meta: FunctionMetadata):
        self._meta = meta
        functools.update_wrapper(self, meta.fn)

    async def __call__(self, *args, **kwargs):
        return await Stent._call_durable_stub(self._meta, args, kwargs)

    async def map(self, iterable) -> list[Any]:
        """Dispatch this function for each item in *iterable* (single arg per call).

        ``await square.map([1, 2, 3])``  ➜  ``[square(1), square(2), square(3)]``
        """
        return await self._map_internal(((item,), {}) for item in iterable)

    async def starmap(self, iterable) -> list[Any]:
        """Dispatch this function for each entry in *iterable*, unpacking args.

        Each entry can be:
        - a tuple of positional args:  ``(a, b)``  ➜  ``fn(a, b)``
        - a (tuple, dict) pair:        ``((a,), {"k": v})``  ➜  ``fn(a, k=v)``

        ``await send_email.starmap([("alice", "hi"), ("bob", "hey")])``
        """
        def _normalise(entry):
            if isinstance(entry, dict):
                return ((), entry)
            if isinstance(entry, (list, tuple)) and len(entry) == 2 and isinstance(entry[1], dict):
                return (tuple(entry[0]), entry[1])
            return (tuple(entry) if isinstance(entry, (list, tuple)) else (entry,), {})
        return await self._map_internal(_normalise(e) for e in iterable)

    async def _map_internal(self, args_iter) -> list[Any]:
        executor: Stent | None = current_executor.get()
        if not executor:
            executor = Stent._default_executor
        if not executor:
            return await asyncio.gather(*[self._meta.fn(*a, **kw) for a, kw in args_iter])

        return await execute_map_batch(
            executor=executor,
            meta=self._meta,
            args_list=list(args_iter),
            exec_id=current_execution_id.get() or "",
            parent_id=current_task_id.get() or "",
            permit_holder=current_permit_holder.get(),
        )


current_execution_id: ContextVar[str | None] = ContextVar("stent_execution_id", default=None)
current_task_id: ContextVar[str | None] = ContextVar("stent_task_id", default=None)
current_worker_semaphore: ContextVar[asyncio.Semaphore | None] = ContextVar("stent_worker_semaphore", default=None)
current_permit_holder: ContextVar[PermitHolder | None] = ContextVar("stent_permit_holder", default=None)
current_worker_id: ContextVar[str | None] = ContextVar("stent_worker_id", default=None)

class Stent:
    backends = Backends
    notifications = Notifications
    default_registry: FunctionRegistry = registry
    _default_executor: ClassVar[Optional[Stent]] = None

    @classmethod
    def use(
        cls,
        backend: Backend,
        serializer: Serializer | Literal["json", "pickle"] = "json",
        notification_backend: NotificationBackend | None = None,
        poll_min_interval: float = 0.1,
        poll_max_interval: float = 5.0,
        poll_backoff_factor: float = 2.0,
        function_registry: FunctionRegistry | None = None,
        metrics: MetricsRecorder | None = None,
    ) -> Stent:
        """Configure a default executor for auto-dispatch.

        After calling this, any ``@Stent.durable()`` function called outside
        of a worker context will automatically dispatch through this executor
        and wait for the result.
        """
        instance = cls(
            backend=backend,
            serializer=serializer,
            notification_backend=notification_backend,
            poll_min_interval=poll_min_interval,
            poll_max_interval=poll_max_interval,
            poll_backoff_factor=poll_backoff_factor,
            function_registry=function_registry,
            metrics=metrics,
        )
        cls._default_executor = instance
        return instance

    @classmethod
    def reset(cls):
        """Clear the default executor (useful for test teardown)."""
        cls._default_executor = None

    def __init__(
        self,
        backend: Backend,
        serializer: Serializer | Literal["json", "pickle"] = "json",
        notification_backend: NotificationBackend | None = None,
        poll_min_interval: float = 0.1,
        poll_max_interval: float = 5.0,
        poll_backoff_factor: float = 2.0,
        function_registry: FunctionRegistry | None = None,
        metrics: MetricsRecorder | None = None,
    ):
        self.backend = backend
        self.registry = function_registry or self.default_registry
        self.serializer: Serializer

        if isinstance(serializer, str):
            if serializer == "json":
                self.serializer = JsonSerializer()
            else:
                 # Local import to avoid circular dependency if pickle serializer was here
                from stent.utils.serialization import PickleSerializer
                self.serializer = PickleSerializer()
        else:
            self.serializer = serializer
            
        self.notification_backend = notification_backend
        self.poll_min_interval = max(0.001, poll_min_interval)
        self.poll_max_interval = max(self.poll_min_interval, poll_max_interval)
        self.poll_backoff_factor = poll_backoff_factor if poll_backoff_factor >= 1.0 else 1.0
        self.metrics: MetricsRecorder = metrics or NoOpMetricsRecorder()
        
        # Tracks background asyncio.Task instances spawned by this executor so their
        # lifecycle can be managed (for example, awaiting or cleanup on shutdown).
        self.background_tasks: set[asyncio.Task[Any]] = set()
        self._active_worker_lifecycles: set[WorkerLifecycle] = set()
        self._default_worker_lifecycle: WorkerLifecycle | None = None
        
        self._register_builtin_tasks()

    def _register_builtin_tasks(self):
        if self.registry.get("stent.sleep"):
            return

        async def sleep_impl(duration: float):
            pass
            
        meta = FunctionMetadata(
            name="stent.sleep",
            fn=sleep_impl,
            cached=False,
            retry_policy=RetryPolicy(),
            tags=["builtin"],
            priority=0,
            queue=None,
            idempotent=True,
            idempotency_key_func=None,
            version="1.0"
        )
        self.registry.register(meta)

    def create_worker_lifecycle(self, *, name: str | None = None) -> WorkerLifecycle:
        """
        Returns a WorkerLifecycle handle that can be passed into serve() so the
        caller can coordinate readiness / draining signals with their host.
        """
        return WorkerLifecycle(name=name)

    def active_worker_lifecycles(self) -> List[WorkerLifecycle]:
        return list(self._active_worker_lifecycles)

    def worker_status_overview(self) -> dict[str, Any]:
        """
        Returns a summary that external health checks can expose.
        """
        handles = self.active_worker_lifecycles()
        ready = any(h.state in ("ready", "draining") and h.ready_event.is_set() for h in handles)
        draining = any(h.state == "draining" for h in handles)
        return {
            "ready": ready,
            "draining": draining,
            "workers": [
                {
                    "name": handle.name or f"worker-{idx}",
                    "state": handle.state,
                }
                for idx, handle in enumerate(handles)
            ],
        }

    def request_worker_drain(self, lifecycle: WorkerLifecycle | None = None) -> None:
        """
        Signals all active workers (or a specific lifecycle) to stop accepting
        new work while letting in-flight tasks finish.
        """
        if lifecycle:
            lifecycle.request_drain()
            return

        for handle in list(self._active_worker_lifecycles):
            handle.request_drain()
        if not self._active_worker_lifecycles and self._default_worker_lifecycle:
            self._default_worker_lifecycle.request_drain()

    async def schedule(
        self, 
        delay: str | dict | timedelta, 
        fn: Callable[..., Awaitable[Any]], 
        *args, 
        **kwargs
    ) -> str:
        """
        Schedule a function to run after a specific delay.
        """
        return await self.dispatch(fn, *args, delay=delay, **kwargs)

    @staticmethod
    async def sleep(duration: str | dict | timedelta):
        """
        Static helper to sleep for a specific duration.
        Proxies to the global sleep function which uses the current executor context.
        """
        # Local import or direct call to global sleep if available in scope
        # Since 'sleep' is defined at the end of this file, we can't call it directly if it's not defined yet.
        # But methods are bound at runtime.
        # However, 'sleep' is defined AFTER 'Stent' class.
        # We can use 'current_executor' directly here.
        executor = current_executor.get()
        if executor:
            await executor.sleep_instance(duration)
        else:
            d = parse_duration(duration)
            await _original_sleep(d.total_seconds())

    async def sleep_instance(self, duration: str | dict | timedelta | int | float):
        """
        Sleep for a specific duration.
        This releases the worker to process other tasks while waiting.
        """
        if isinstance(duration, (int, float)):
            d = timedelta(seconds=duration)
        else:
            d = parse_duration(duration)
        meta = self.registry.get("stent.sleep")
        if not meta:
            self._register_builtin_tasks()
            meta = self.registry.get("stent.sleep")
        if not meta:
            raise RuntimeError("stent.sleep not registered")
        
        # Schedule the sleep task to run AFTER the duration.
        # The orchestrator will wait for it to complete.
        await self._schedule_activity(meta, (d.total_seconds(),), {}, None, delay=d)

    @classmethod
    def durable(
        cls,
        *,
        cached: bool = False,
        retry_policy: RetryPolicy | None = None,
        tags: List[str] | None = None,
        priority: int = 0,
        queue: str | None = None,
        idempotent: bool = False,
        idempotency_key_func: Callable[..., str] | None = None,
        version: str | None = None,
        max_concurrent: int | None = None,
    ):
        def decorator(fn):
            reg = cls.default_registry
            name = reg.name_for_function(fn)
            meta = FunctionMetadata(
                name=name,
                fn=fn,
                cached=cached,
                retry_policy=retry_policy or RetryPolicy(),
                tags=tags or [],
                priority=priority,
                queue=queue,
                idempotent=idempotent,
                idempotency_key_func=idempotency_key_func,
                version=version,
                max_concurrent=max_concurrent,
            )
            reg.register(meta)
            return DurableFunction(meta)
        return decorator

    @classmethod
    async def wrap(cls, fn: Callable[..., Awaitable[Any]], args: tuple, kwargs: dict | None = None):
         if kwargs is None:
             kwargs = {}

         executor = current_executor.get()
         reg = executor.registry if executor else cls.default_registry
         name = reg.name_for_function(fn)
         meta = reg.get(name)
         if not meta:
             raise UnregisteredFunctionError(name)

         return await cls._call_durable_stub(meta, args, kwargs)

    @classmethod
    async def map(
        cls,
        fn: Callable[..., Awaitable[Any]],
        iterable: Any
    ) -> List[Any]:
        """Batch-dispatch *fn* for each item in *iterable*.

        Prefer ``fn.map(items)`` instead.  This classmethod is kept for
        backwards compatibility.
        """
        if isinstance(fn, DurableFunction):
            return await fn.map(iterable)

        executor: Optional[Stent] = current_executor.get()
        if not executor:
            return await asyncio.gather(*[fn(item) for item in iterable])

        reg = executor.registry
        name = reg.name_for_function(fn)
        meta = reg.get(name)
        if not meta:
            raise UnregisteredFunctionError(name)

        return await execute_map_batch(
            executor=executor,
            meta=meta,
            args_list=[((item,), {}) for item in iterable],
            exec_id=current_execution_id.get() or "",
            parent_id=current_task_id.get() or "",
            permit_holder=current_permit_holder.get(),
        )

    @classmethod
    async def gather(cls, *tasks, **kwargs):
        """
        Alias for asyncio.gather. 
        Note: If you pass function calls (e.g. `stent.gather(func(1), func(2))`), 
        they are scheduled immediately when called, not batched by gather.
        Use `stent.map` for batch scheduling optimization if applicable.
        
        Supports `return_exceptions=True`.
        """
        return await asyncio.gather(*tasks, **kwargs)

    @classmethod
    def context(
        cls,
        *,
        counters: dict[str, int | float] | None = None,
        state: dict[str, Any] | None = None,
    ) -> ExecutionContext:
        executor = current_executor.get()
        if not executor:
            raise RuntimeError("Cannot access context outside of durable function")

        exec_id = current_execution_id.get()
        if not exec_id:
            raise RuntimeError("Cannot access context without active execution")

        ctx = current_execution_context.get()
        if ctx is None:
            ctx = ExecutionContext(executor, exec_id)
            current_execution_context.set(ctx)

        if counters or state:
            ctx._track(ctx.initialize(counters=counters, state=state))
        return ctx

    @classmethod
    async def _call_durable_stub(cls, meta: FunctionMetadata, args: tuple, kwargs: dict):
        executor = current_executor.get() # Get the current executor if any
        exec_id = current_execution_id.get()
        
        # If no executor in context, check for a default executor for auto-dispatch.
        # If neither exists, fall through to a direct local call (unit test / no-runtime mode).
        if not executor:
            executor = cls._default_executor
            if not executor:
                logger.debug(f"STUB: {meta.name} running locally (no executor context)")
                return await meta.fn(*args, **kwargs)

            # Auto-dispatch: create a full execution and wait for a worker to process it.
            exec_id = await executor.dispatch(meta.fn, *args, **kwargs)
            result = await executor.wait_for(exec_id)
            if result.ok:
                return result.value
            if isinstance(result.error, BaseException):
                raise result.error
            raise Exception(str(result.error))

        # Now we know we are within an executor's context (i.e., this is a distributed call)

        # Idempotency / Caching check - applies to any durable function call within an execution
        # This key is generated here and passed down if not a hit.
        key = None
        if meta.idempotent or meta.cached:
             if meta.idempotency_key_func:
                 key = meta.idempotency_key_func(*args, **kwargs)
             else:
                 key = default_idempotency_key(meta.name, meta.version, args, kwargs, serializer=executor.serializer)

        if key:
            if meta.cached:
                 cached_val = await executor.backend.get_cached_result(key)
                 logger.debug(f"_call_durable_stub: fetched cached_val: {cached_val is not None} for key {key}")
                 logger.debug(f"_call_durable_stub: type(cached_val)={type(cached_val)}, cached_val={cached_val}, bool(cached_val)={bool(cached_val)}")
                 should_hit_cache = (cached_val is not None)
                 logger.debug(f"_call_durable_stub: should_hit_cache={should_hit_cache}")
                 if should_hit_cache:
                      logger.info(f"Cache HIT for {meta.name} with key {key}")
                      if exec_id: # Only append progress if we are in an execution
                          await executor.backend.append_progress(exec_id, ExecutionProgress(
                              step=meta.name, status="cache_hit"
                          ))
                      return executor.serializer.loads(cached_val)

            # This 'if' should be at the same indentation level as 'if meta.cached:'
            if meta.idempotent:
                stored_val = await executor.backend.get_idempotency_result(key)
                if stored_val is not None:
                    logger.info(f"Idempotency HIT for {meta.name} with key {key}")
                    if exec_id: # Only append progress if we are in an execution
                        await executor.backend.append_progress(exec_id, ExecutionProgress(
                            step=meta.name, status="cache_hit" # Using cache_hit status for idempotency too
                        ))
                    return executor.serializer.loads(stored_val)

        # If we reached here, it's not a cache/idempotency hit, so schedule as an activity.
        # The key is passed so _schedule_activity can store it in the TaskRecord.
        completed_task = await executor._schedule_activity(meta, args, kwargs, key)
        
        if completed_task.state == "failed":
            # Propagate error from activity
            if completed_task.error:
                err = executor.serializer.loads(completed_task.error)
                if isinstance(err, dict) and "__type__" in err: 
                    raise Exception(str(err))
                if isinstance(err, BaseException):
                    raise err
                raise Exception(str(err))
            raise Exception("Task failed without error info")

        completed_task_result = completed_task.result or b"null"
        res = executor.serializer.loads(completed_task_result)
        
        # Store cache/idempotency if enabled and key was generated after successful execution
        if key:
            if meta.cached:
                await executor.backend.set_cached_result(key, completed_task_result)
                logger.debug(f"Stored cached result for {meta.name} with key {key}")
            if meta.idempotent:
                await executor.backend.set_idempotency_result(key, completed_task_result)
                logger.debug(f"Stored idempotency result for {meta.name} with key {key}")
            
        return res

    async def dispatch(
        self,
        fn: Callable[..., Awaitable[Any]],
        *args,
        expiry: str | timedelta | None = None,
        max_duration: str | timedelta | None = None,
        delay: str | dict | timedelta | None = None,
        tags: List[str] | None = None,
        priority: int = 0,
        queue: str | None = None,
        **kwargs,
    ) -> str:
        reg = self.registry
        name = reg.name_for_function(fn)
        meta = reg.get(name)
        if not meta:
            raise UnregisteredFunctionError(name)

        normalized_expiry, normalized_delay = normalize_dispatch_timing(
            expiry=expiry,
            max_duration=max_duration,
            delay=delay,
        )

        execution_id, record, task = build_dispatch_records(
            name=name,
            args_bytes=self.serializer.dumps(args),
            kwargs_bytes=self.serializer.dumps(kwargs),
            retry_policy=meta.retry_policy,
            tags=tags or meta.tags,
            priority=priority,
            queue=queue or meta.queue,
            expiry=normalized_expiry,
            delay=normalized_delay,
        )
        await persist_dispatch_records(executor=self, execution=record, root_task=task)

        return execution_id

    async def send_signal(self, execution_id: str, name: str, payload: Any) -> None:
        """
        Send a signal to a running execution. 
        If the execution is waiting for this signal, it will be resumed.
        If not, the signal will be buffered.
        """
        payload_bytes = self.serializer.dumps(payload)
        await persist_signal_and_wake_waiter(
            execution_id=execution_id,
            name=name,
            payload_bytes=payload_bytes,
            create_signal=self.backend.create_signal,
            list_tasks_for_execution=self.backend.list_tasks_for_execution,
            update_task=self.backend.update_task,
            notify_task_completed=(
                self.notification_backend.notify_task_completed
                if self.notification_backend
                else None
            ),
        )

    @staticmethod
    async def wait_for_signal(name: str) -> Any:
        """
        Static helper to wait for a signal.
        Proxies to the global executor context.
        """
        executor = current_executor.get()
        if not executor:
             raise Exception("Cannot wait for signal outside of durable function")
        return await executor.wait_for_signal_instance(name)

    async def wait_for_signal_instance(self, name: str) -> Any:
        """
        Pauses the current workflow until a signal with the given name is received.
        """
        exec_id = current_execution_id.get()
        if not exec_id:
             raise Exception("Cannot wait for signal outside of durable function")
             
        return await resolve_signal_wait(
            execution_id=exec_id,
            name=name,
            parent_task_id=current_task_id.get(),
            loads=self.serializer.loads,
            get_task=self.backend.get_task,
            get_signal=self.backend.get_signal,
            create_signal=self.backend.create_signal,
            create_task=self.backend.create_task,
            append_progress=self.backend.append_progress,
            wait_for_task=self._wait_for_task,
        )

    async def _schedule_activity(
        self, 
        meta: FunctionMetadata, 

        args: tuple, 
        kwargs: dict, 
        idempotency_key: str | None,
        delay: timedelta | None = None
    ) -> TaskRecord:
        return await schedule_activity_and_wait(
            executor=self,
            meta=meta,
            args=args,
            kwargs=kwargs,
            idempotency_key=idempotency_key,
            delay=delay,
            exec_id=current_execution_id.get() or "",
            parent_id=current_task_id.get() or "",
        )

    async def _wait_for_task_internal(self, task_id: str, expiry: float | None = None) -> TaskRecord:
        """Waits for task completion without modifying semaphore."""
        return await wait_for_task_terminal(
            task_id=task_id,
            get_task=self.backend.get_task,
            notification_backend=self.notification_backend,
            expiry=expiry,
            poll_min_interval=self.poll_min_interval,
            poll_max_interval=self.poll_max_interval,
            poll_backoff_factor=self.poll_backoff_factor,
            sleep_fn=_original_sleep,
            expiry_error_factory=ExpiryError,
        )

    async def _wait_for_task(self, task_id: str, expiry: float | None = None) -> TaskRecord:
        permit = current_permit_holder.get()
        if permit:
            permit.release()
            
        try:
            return await self._wait_for_task_internal(task_id, expiry)
        finally:
            if permit:
                await permit.acquire()

    async def state_of(self, execution_id: str) -> ExecutionState:
        record = await self.backend.get_execution(execution_id)
        if not record:
            raise ValueError("Execution not found")

        counters_method = getattr(self.backend, "get_execution_counters", None)
        if callable(counters_method):
            counters = await cast(Awaitable[dict[str, int | float]], counters_method(execution_id))
        else:
            counters = {}

        state_method = getattr(self.backend, "get_execution_state_values", None)
        if callable(state_method):
            state_values = await cast(Awaitable[dict[str, bytes]], state_method(execution_id))
        else:
            state_values = {}
        custom_state = {
            key: self.serializer.loads(value)
            for key, value in state_values.items()
        }
        # Optional: refresh progress?
        return ExecutionState(
            id=record.id,
            state=record.state,
            result=self.serializer.loads(record.result) if record.result else None,
            started_at=record.started_at,
            completed_at=record.completed_at,
            retries=record.retries,
            progress=record.progress,
            tags=record.tags,
            priority=record.priority,
            queue=record.queue,
            counters=counters,
            custom_state=custom_state,
        )

    async def result_of(self, execution_id: str) -> Result[Any, Any]:
        record = await self.backend.get_execution(execution_id)
        if not record:
            raise ValueError("Execution not found")
        if record.state not in ("completed", "failed", "timed_out", "cancelled"):
             raise Exception("Execution still running")
        
        if record.result:
            val = self.serializer.loads(record.result)
            if isinstance(val, Result):
                return val
            return Result.Ok(val)
        if record.error:
            # Wrap error in Result.Error if it's raw exception
            err = self.serializer.loads(record.error)
            if isinstance(err, Result):
                 return err
            return Result.Error(err)

        if record.state == "cancelled":
            return Result.Error(Exception("Execution cancelled"))
             
        raise Exception("No result available")

    async def wait_for(self, execution_id: str, expiry: float | None = None) -> Result[Any, Any]:
        """
        Blocks until the execution with the given ID is completed, failed, or timed out.
        Returns the result of the execution.
        """
        # Quick check first
        try:
            return await self.result_of(execution_id)
        except Exception:
            pass # Not done yet

        await wait_for_execution_terminal(
            execution_id=execution_id,
            state_of=self.state_of,
            notification_backend=self.notification_backend,
            expiry=expiry,
            sleep_interval=0.5,
            expiry_error_factory=ExpiryError,
        )

        return await self.result_of(execution_id)

    async def list_executions(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[ExecutionState]:
        """
        List executions with optional filtering by state.
        Returns a list of ExecutionState objects (without full progress history for efficiency).
        """
        records = await self.backend.list_executions(limit, offset, state)
        return [
            ExecutionState(
                id=r.id,
                state=r.state,
                result=None, # Don't deserialize result for summary
                started_at=r.started_at,
                completed_at=r.completed_at,
                retries=r.retries,
                progress=[], # Skip progress
                tags=r.tags,
                priority=r.priority,
                queue=r.queue
            ) for r in records
        ]

    async def queue_depth(self, queue: str | None = None) -> int:
        """
        Returns the number of pending tasks in the specified queue (or all queues if None).
        """
        return await self.backend.count_tasks(queue=queue, state="pending")

    async def get_running_activities(self) -> List[TaskRecord]:
        """
        Returns a list of currently running tasks (activities).
        """
        return await self.backend.list_tasks(limit=100, state="running")

    async def list_dead_letters(self, limit: int = 50) -> List[DeadLetterRecord]:
        """
        Returns the most recent dead-lettered tasks.
        """
        return await self.backend.list_dead_tasks(limit=limit)

    async def get_dead_letter(self, task_id: str) -> DeadLetterRecord | None:
        return await self.backend.get_dead_task(task_id)

    async def discard_dead_letter(self, task_id: str) -> bool:
        return await self.backend.delete_dead_task(task_id)

    async def replay_dead_letter(
        self,
        task_id: str,
        *,
        queue: str | None = None,
        reset_retries: bool = True,
    ) -> str:
        """
        Re-enqueues a dead-lettered task with a fresh task ID so it can run again.
        """
        record = await self.get_dead_letter(task_id)
        if not record:
            raise ValueError(f"Dead-letter task {task_id} not found")

        new_task = replace(record.task)
        original_id = new_task.id
        new_task.id = str(uuid.uuid4())
        new_task.state = "pending"
        new_task.worker_id = None
        new_task.lease_expires_at = None
        new_task.started_at = None
        new_task.completed_at = None
        new_task.result = None
        new_task.error = None
        new_task.created_at = datetime.now()
        if reset_retries:
            new_task.retries = 0
        if queue is not None:
            new_task.queue = queue

        await self.backend.create_task(new_task)
        await self.backend.delete_dead_task(original_id)

        if record.task.kind == "orchestrator":
            execution = await self.backend.get_execution(new_task.execution_id)
            if execution:
                execution.state = "pending"
                execution.started_at = None
                execution.completed_at = None
                execution.result = None
                execution.error = None
                execution.retries = 0
                await self.backend.update_execution(execution)

        return new_task.id

    async def serve(
        self,
        *,
        worker_id: str | None = None,
        queues: List[str] | None = None,
        tags: List[str] | None = None,
        max_concurrency: int = 10,
        lease_duration: timedelta = timedelta(minutes=5),
        heartbeat_interval: timedelta | None = None,
        poll_interval: float = 1.0,
        poll_interval_max: float | None = None,
        poll_backoff_factor: float | None = None,
        cleanup_interval: float | None = 3600.0, # Default 1 hour
        retention_period: timedelta = timedelta(days=7),
        lifecycle: WorkerLifecycle | None = None,
    ):
        if not worker_id:
            worker_id = str(uuid.uuid4())

        lifecycle_handle = lifecycle or self.create_worker_lifecycle(name=worker_id)
        if lifecycle is None:
            self._default_worker_lifecycle = lifecycle_handle
        lifecycle_handle.reset()
        self._active_worker_lifecycles.add(lifecycle_handle)
        worker_tasks: set[asyncio.Task[Any]] = set()

        if heartbeat_interval is None and lease_duration > timedelta(0):
            heartbeat_interval = lease_duration / 2
            min_interval = timedelta(milliseconds=100)
            if heartbeat_interval < min_interval:
                heartbeat_interval = min_interval
        elif heartbeat_interval is not None and heartbeat_interval <= timedelta(0):
            heartbeat_interval = None
            
        worker_poll_min = max(0.001, poll_interval)
        worker_poll_max = poll_interval_max if poll_interval_max is not None else self.poll_max_interval
        worker_poll_max = max(worker_poll_min, worker_poll_max)
        worker_poll_backoff = (
            poll_backoff_factor if poll_backoff_factor is not None and poll_backoff_factor >= 1.0
            else self.poll_backoff_factor
        )
        current_poll_delay = worker_poll_min

        sem = asyncio.Semaphore(max_concurrency)
        logger.info(f"Worker {worker_id} started. Queues: {queues}")
        lifecycle_handle.mark_ready()
        
        cleanup_task = None
        if cleanup_interval is not None:
            cleanup_task = asyncio.create_task(self._cleanup_loop(retention_period, cleanup_interval))

        try:
            while True:
                if lifecycle_handle.stop_event.is_set():
                    lifecycle_handle.mark_draining()
                    if worker_tasks:
                        await asyncio.wait(worker_tasks, return_when=asyncio.FIRST_COMPLETED)
                        continue
                    break

                # Check if we can acquire
                # We don't want to block here, just check if sem is full
                # But asyncio.Semaphore doesn't have a check without acquire. 
                # We acquire first, then claim.
                await sem.acquire()
                if lifecycle_handle.stop_event.is_set():
                    lifecycle_handle.mark_draining()
                    sem.release()
                    continue

                # Collect concurrency limits
                concurrency_limits = {}
                for name, meta in self.registry.items():
                    if meta.max_concurrent is not None:
                        concurrency_limits[name] = meta.max_concurrent

                try:
                    task = await self.backend.claim_next_task(
                        worker_id=worker_id,
                        queues=queues,
                        tags=tags,
                        now=datetime.now(),
                        lease_duration=lease_duration,
                        concurrency_limits=concurrency_limits
                    )
                    
                    if task:
                        # logger.info(f"Worker claimed task {task.step_name} ({task.id})")
                        self.metrics.task_claimed(queue=task.queue, step_name=task.step_name, kind=task.kind)
                        # Run in background
                        bg_task = asyncio.create_task(
                            self._handle_task(task, worker_id, lease_duration, heartbeat_interval, sem)
                        )
                        self.background_tasks.add(bg_task)
                        worker_tasks.add(bg_task)
                        bg_task.add_done_callback(self.background_tasks.discard)
                        bg_task.add_done_callback(worker_tasks.discard)
                        current_poll_delay = worker_poll_min
                    else:
                        sem.release()
                        await _original_sleep(current_poll_delay)
                        current_poll_delay = min(
                            worker_poll_max, current_poll_delay * worker_poll_backoff
                        )
                except Exception as e:
                    sem.release()
                    logger.error(f"Error in worker loop: {e}")
                    await _original_sleep(current_poll_delay)
                    current_poll_delay = min(
                        worker_poll_max, current_poll_delay * worker_poll_backoff
                    )
        except asyncio.CancelledError:
            logger.info("Worker cancelled")
            raise
        finally:
            if cleanup_task:
                cleanup_task.cancel()
                try:
                    await cleanup_task
                except asyncio.CancelledError:
                    pass
            lifecycle_handle.mark_stopped()
            self._active_worker_lifecycles.discard(lifecycle_handle)

    async def shutdown(self):
        """
        Gracefully shuts down the executor, waiting for pending background tasks.
        """
        if self.background_tasks:
            logger.info(f"Waiting for {len(self.background_tasks)} background tasks to complete...")
            await asyncio.gather(*self.background_tasks, return_exceptions=True)

    async def _cleanup_loop(self, retention: timedelta, interval: float):
        logger.info(f"Starting cleanup loop. Retention: {retention}, Interval: {interval}s")
        while True:
            try:
                # Jitter the startup/interval slightly to avoid thundering herd if multiple workers start at once
                # But simple sleep is fine for now
                await _original_sleep(interval)
                
                cutoff = datetime.now() - retention
                count = await self.backend.cleanup_executions(cutoff)
                if count > 0:
                    logger.info(f"Cleaned up {count} old executions (older than {cutoff})")
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Cleanup failed: {e}")
                await _original_sleep(60) # Wait a bit before retrying on error

    async def _lease_heartbeat_loop(
        self,
        *,
        task_id: str,
        worker_id: str,
        lease_duration: timedelta,
        interval: timedelta,
        stop_event: asyncio.Event,
    ) -> None:
        await lease_heartbeat_loop(
            executor=self,
            task_id=task_id,
            worker_id=worker_id,
            lease_duration=lease_duration,
            interval=interval,
            stop_event=stop_event,
            logger=logger,
        )

    async def _handle_task(
        self, 
        task: TaskRecord, 
        worker_id: str, 
        lease_duration: timedelta, 
        heartbeat_interval: timedelta | None,
        sem: asyncio.Semaphore
    ):
        # logger.info(f"DEBUG: Starting _handle_task for {task.step_name} ({task.id})")
        token_exec = current_execution_id.set(task.execution_id)
        token_task = current_task_id.set(task.id)
        token_executor = current_executor.set(self)
        token_sem = current_worker_semaphore.set(sem)
        token_permit = current_permit_holder.set(PermitHolder(sem))
        token_worker = current_worker_id.set(worker_id)

        try:
            await run_claimed_task(
                executor=self,
                task=task,
                worker_id=worker_id,
                lease_duration=lease_duration,
                heartbeat_interval=heartbeat_interval,
                expiry_error_type=ExpiryError,
                unregistered_error_factory=UnregisteredFunctionError,
                logger=logger,
            )
        finally:
            sem.release()
            current_execution_id.reset(token_exec)
            current_task_id.reset(token_task)
            current_executor.reset(token_executor)
            current_worker_semaphore.reset(token_sem)
            current_permit_holder.reset(token_permit)
            current_worker_id.reset(token_worker)
            # logger.info(f"DEBUG: Finished _handle_task for {task.step_name} ({task.id})")

# Context var for executor instance
current_executor: ContextVar[Optional[Stent]] = ContextVar("stent_executor", default=None)



Stent.backends = Backends
Stent.notifications = Notifications

# ---------------------------------------------------------------------------
# Durable sleep threshold (seconds).  Any ``asyncio.sleep(n)`` call made
# inside a durable execution where *n* >= this value is automatically
# promoted to a durable sleep that releases the worker.  Sleeps below
# the threshold use the real ``asyncio.sleep``.
# ---------------------------------------------------------------------------
DURABLE_SLEEP_THRESHOLD: float = 1.0


async def sleep(duration: str | dict | timedelta):
    """
    Global sleep helper that uses the current executor context if available,
    otherwise falls back to asyncio.sleep (non-durable).
    """
    executor = current_executor.get()
    if executor:
        await executor.sleep(duration)
    else:
        # Fallback for local testing or non-durable usage
        d = parse_duration(duration)
        await _original_sleep(d.total_seconds())


async def _durable_sleep_wrapper(delay, result=None, *, _real_sleep=_original_sleep):
    """Drop-in replacement for ``asyncio.sleep`` that promotes long sleeps
    to durable sleeps when running inside a Stent execution context."""
    executor: Stent | None = current_executor.get()
    if executor and isinstance(delay, (int, float)) and delay >= DURABLE_SLEEP_THRESHOLD:
        await executor.sleep_instance(delay)
        return result
    return await _real_sleep(delay, result)


# Patch asyncio.sleep so that user code written with the standard library
# automatically benefits from durable sleep inside Stent workers.
asyncio.sleep = _durable_sleep_wrapper  # type: ignore[assignment]



from stent.executor import Stent, DurableFunction, sleep, WorkerLifecycle, install_structured_logging
from stent.core import Result, RetryPolicy, ExecutionState
from stent.registry import registry
from stent.metrics import MetricsRecorder

__all__ = [
    "Stent",
    "DurableFunction",
    "Result",
    "RetryPolicy",
    "ExecutionState",
    "registry",
    "sleep",
    "WorkerLifecycle",
    "install_structured_logging",
    "MetricsRecorder",
]



import sys
from datetime import datetime
from typing import Any, cast

import pytest

import stent.cli as cli
from stent.core import ExecutionState, Result


class _SpyBackend:
    def __init__(self, *, raise_on_list: bool = False):
        self.raise_on_list = raise_on_list
        self.init_calls = 0
        self.close_calls = 0

    async def init_db(self) -> None:
        self.init_calls += 1

    async def close(self) -> None:
        self.close_calls += 1

    async def list_executions(self, limit: int = 10, offset: int = 0, state: str | None = None):
        if self.raise_on_list:
            raise RuntimeError("boom")
        return []


class _StatsBackend(_SpyBackend):
    def __init__(self):
        super().__init__()
        self.count_executions_calls = 0
        self.count_dead_tasks_calls = 0

    async def count_executions(self, state: str | None = None) -> int:
        self.count_executions_calls += 1
        return 0

    async def count_tasks(self, queue: str | None = None, state: str | None = None) -> int:
        return 0

    async def list_tasks(self, limit: int = 10, offset: int = 0, state: str | None = None):
        return []

    async def count_dead_tasks(self) -> int:
        self.count_dead_tasks_calls += 1
        return 0

    async def list_dead_tasks(self, limit: int = 50):
        raise AssertionError("stats should use count_dead_tasks, not list_dead_tasks")

    async def get_task(self, task_id: str):
        return None

    async def get_execution(self, execution_id: str):
        return None


class _ShowExecutor:
    async def state_of(self, _execution_id: str) -> ExecutionState:
        return ExecutionState(
            id="exec-1",
            state="running",
            result=Result.Ok("ok"),
            started_at=datetime.now(),
            completed_at=None,
            retries=0,
            progress=[],
            tags=["alpha"],
            priority=1,
            queue=None,
            counters={"progress": 1, "volume": 30, "abc": 40},
            custom_state={"owner": "foo", "phase": "bar_done"},
        )


@pytest.mark.asyncio
async def test_cli_initializes_and_closes_backend(monkeypatch):
    backend = _SpyBackend()

    monkeypatch.setattr(cli.Stent.backends, "SQLiteBackend", staticmethod(lambda _db: backend))
    monkeypatch.setattr(sys, "argv", ["stent", "list"])

    result = await cli.main_async()
    assert result == 0
    assert backend.init_calls == 1
    assert backend.close_calls == 1


@pytest.mark.asyncio
async def test_cli_closes_backend_when_command_errors(monkeypatch):
    backend = _SpyBackend(raise_on_list=True)

    monkeypatch.setattr(cli.Stent.backends, "SQLiteBackend", staticmethod(lambda _db: backend))
    monkeypatch.setattr(sys, "argv", ["stent", "list"])

    with pytest.raises(RuntimeError, match="boom"):
        await cli.main_async()

    assert backend.init_calls == 1
    assert backend.close_calls == 1


@pytest.mark.asyncio
async def test_cli_stats_uses_count_apis(monkeypatch):
    backend = _StatsBackend()

    async def _forbidden_list_executions(limit: int = 10, offset: int = 0, state: str | None = None):
        raise AssertionError("stats should use count_executions, not list_executions")

    backend.list_executions = _forbidden_list_executions  # type: ignore[method-assign]

    monkeypatch.setattr(cli.Stent.backends, "SQLiteBackend", staticmethod(lambda _db: backend))
    monkeypatch.setattr(sys, "argv", ["stent", "stats"])

    result = await cli.main_async()

    assert result == 0
    assert backend.init_calls == 1
    assert backend.close_calls == 1
    assert backend.count_executions_calls > 0
    assert backend.count_dead_tasks_calls > 0


@pytest.mark.asyncio
async def test_show_execution_prints_counters_and_custom_state(capsys):
    class _Args:
        id = "exec-1"

    result = await cli.show_execution(cast(Any, _ShowExecutor()), _Args())
    out = capsys.readouterr().out

    assert result == 0
    assert "Counters" in out
    assert "progress: 1" in out
    assert "volume: 30" in out
    assert "abc: 40" in out
    assert "Custom State" in out
    assert "owner: foo" in out
    assert "phase: bar_done" in out



from __future__ import annotations

import uuid
from datetime import datetime
from typing import Any, Awaitable, Callable

from stent.core import ExecutionProgress, SignalRecord, TaskRecord


def signal_step_name(name: str) -> str:
    return f"signal:{name}"


def deterministic_signal_task_id(execution_id: str, step_name: str) -> str:
    try:
        exec_uuid = uuid.UUID(execution_id)
    except ValueError:
        exec_uuid = uuid.uuid4()
    return str(uuid.uuid5(exec_uuid, step_name))


async def persist_signal_and_wake_waiter(
    *,
    execution_id: str,
    name: str,
    payload_bytes: bytes,
    create_signal: Callable[[SignalRecord], Awaitable[None]],
    list_tasks_for_execution: Callable[[str], Awaitable[list[TaskRecord]]],
    update_task: Callable[[TaskRecord], Awaitable[None]],
    notify_task_completed: Callable[[str], Awaitable[None]] | None,
) -> None:
    signal = SignalRecord(
        execution_id=execution_id,
        name=name,
        payload=payload_bytes,
        created_at=datetime.now(),
        consumed=False,
    )
    await create_signal(signal)

    target_step = signal_step_name(name)
    tasks = await list_tasks_for_execution(execution_id)
    for task in tasks:
        if task.step_name == target_step and task.kind == "signal" and task.state == "pending":
            task.state = "completed"
            task.result = payload_bytes
            task.completed_at = datetime.now()
            await update_task(task)
            if notify_task_completed:
                await notify_task_completed(task.id)
            break


async def resolve_signal_wait(
    *,
    execution_id: str,
    name: str,
    parent_task_id: str | None,
    loads: Callable[[bytes], Any],
    get_task: Callable[[str], Awaitable[TaskRecord | None]],
    get_signal: Callable[[str, str], Awaitable[SignalRecord | None]],
    create_signal: Callable[[SignalRecord], Awaitable[None]],
    create_task: Callable[[TaskRecord], Awaitable[None]],
    append_progress: Callable[[str, ExecutionProgress], Awaitable[None]],
    wait_for_task: Callable[[str], Awaitable[TaskRecord]],
) -> Any:
    step_name = signal_step_name(name)
    task_id = deterministic_signal_task_id(execution_id, step_name)

    existing_task = await get_task(task_id)
    if existing_task:
        if existing_task.state == "completed":
            return loads(existing_task.result) if existing_task.result is not None else None
        completed = await wait_for_task(task_id)
        return loads(completed.result) if completed.result is not None else None

    signal = await get_signal(execution_id, name)
    if signal and not signal.consumed:
        signal.consumed = True
        signal.consumed_at = datetime.now()
        await create_signal(signal)

        task = TaskRecord(
            id=task_id,
            execution_id=execution_id,
            step_name=step_name,
            kind="signal",
            parent_task_id=parent_task_id,
            state="completed",
            args=b"",
            kwargs=b"",
            retries=0,
            created_at=datetime.now(),
            completed_at=datetime.now(),
            tags=["signal"],
            priority=0,
            queue=None,
            retry_policy=None,
            result=signal.payload,
        )
        await create_task(task)
        await append_progress(
            execution_id,
            ExecutionProgress(step=step_name, status="completed", detail="Signal consumed from buffer"),
        )
        return loads(signal.payload)

    pending_task = TaskRecord(
        id=task_id,
        execution_id=execution_id,
        step_name=step_name,
        kind="signal",
        parent_task_id=parent_task_id,
        state="pending",
        args=b"",
        kwargs=b"",
        retries=0,
        created_at=datetime.now(),
        tags=["signal"],
        priority=0,
        queue=None,
        retry_policy=None,
    )
    await create_task(pending_task)
    await append_progress(
        execution_id,
        ExecutionProgress(step=step_name, status="dispatched", detail="Waiting for signal"),
    )

    completed = await wait_for_task(task_id)

    latest_signal = await get_signal(execution_id, name)
    if latest_signal and not latest_signal.consumed:
        latest_signal.consumed = True
        latest_signal.consumed_at = datetime.now()
        await create_signal(latest_signal)

    return loads(completed.result) if completed.result is not None else None



import redis.asyncio as redis
import json
import asyncio
from typing import AsyncIterator, Dict, Any, Optional
from stent.notifications.base import NotificationBackend

class RedisBackend(NotificationBackend):
    def __init__(self, url: str):
        self.redis = redis.from_url(url)

    async def notify_task_completed(self, task_id: str) -> None:
        await self.redis.publish(f"stent:task:{task_id}", json.dumps({
            "task_id": task_id,
            "state": "completed"
        }))

    async def notify_task_updated(self, task_id: str, state: str) -> None:
        await self.redis.publish(f"stent:task:{task_id}", json.dumps({
            "task_id": task_id,
            "state": state
        }))

    async def subscribe_to_task(
        self,
        task_id: str,
        *,
        expiry: float | None = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        pubsub = self.redis.pubsub()
        channel = f"stent:task:{task_id}"
        await pubsub.subscribe(channel)
        
        try:
            async def _listen():
                async for message in pubsub.listen():
                    if message["type"] == "message":
                        data = json.loads(message["data"])
                        yield data
                        if data.get("state") in ("completed", "failed"):
                            break

            if expiry:
                try:
                    async with asyncio.timeout(expiry):
                        async for item in _listen():
                            yield item
                except asyncio.TimeoutError:
                    raise
            else:
                 async for item in _listen():
                     yield item
        finally:
            await pubsub.unsubscribe(channel)
            await pubsub.close()

    async def notify_execution_updated(self, execution_id: str, state: str) -> None:
        await self.redis.publish(f"stent:execution:{execution_id}", json.dumps({
            "execution_id": execution_id,
            "state": state
        }))

    async def subscribe_to_execution(
        self,
        execution_id: str,
        *,
        expiry: float | None = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        pubsub = self.redis.pubsub()
        channel = f"stent:execution:{execution_id}"
        await pubsub.subscribe(channel)
        
        try:
            async def _listen():
                async for message in pubsub.listen():
                    if message["type"] == "message":
                        data = json.loads(message["data"])
                        yield data
                        if data.get("state") in ("completed", "failed", "timed_out", "cancelled"):
                            break

            if expiry:
                try:
                    async with asyncio.timeout(expiry):
                        async for item in _listen():
                            yield item
                except asyncio.TimeoutError:
                    raise
            else:
                 async for item in _listen():
                     yield item
        finally:
            await pubsub.unsubscribe(channel)
            await pubsub.close()



import asyncio
import os
import logging
from datetime import timedelta
from stent import Stent, Result, RetryPolicy

logging.basicConfig(level=logging.INFO)

# --- Durable Functions Definition ---

@Stent.durable(retry_policy=RetryPolicy(max_attempts=3, initial_delay=0.1))
async def fetch_external_data(url: str) -> Result[str, Exception]:
    # Simulate external API call that might fail
    print(f"[Task] Fetching data from {url}")
    await asyncio.sleep(0.3)
    if "fail" in url:
        raise ConnectionError(f"Failed to connect to {url}")
    return Result.Ok(f"<data from {url}>")

@Stent.durable(cached=True, version="v1.0")
async def process_data_heavy(raw_data: str) -> Result[str, Exception]:
    # Simulate a CPU-bound or idempotent data processing task
    print(f"[Task] Processing data: {raw_data[:20]}...")
    await asyncio.sleep(0.7)
    processed = raw_data.upper().replace("DATA", "PROCESSED_DATA")
    return Result.Ok(processed)

@Stent.durable(idempotent=True, queue="notifications")
async def send_notification(recipient: str, subject: str, body: str) -> Result[bool, Exception]:
    # Simulate sending an email or other notification
    print(f"[Task] Sending notification to {recipient} with subject: {subject}")
    await asyncio.sleep(0.2)
    if "error" in recipient:
        raise ValueError("Invalid recipient for notification")
    return Result.Ok(True)

@Stent.durable(tags=["orchestrator", "pipeline"])
async def complex_data_pipeline(config: dict) -> Result[dict, Exception]:
    print("[Orchestrator] Starting complex data pipeline")

    # Step 1: Fetch data (can retry)
    url1 = config.get("url1", "http://example.com/data1")
    fetch_result1 = await fetch_external_data(url1)
    if not fetch_result1.ok:
        print(f"[Orchestrator] Failed to fetch data1: {fetch_result1.error}")
        return Result.Error(Exception(f"Pipeline failed at fetch_external_data: {fetch_result1.error}"))
    raw_data1 = fetch_result1.or_raise()

    url2 = config.get("url2", "http://example.com/data2")
    fetch_result2 = await fetch_external_data(url2)
    if not fetch_result2.ok:
        print(f"[Orchestrator] Failed to fetch data2: {fetch_result2.error}")
        return Result.Error(Exception(f"Pipeline failed at fetch_external_data: {fetch_result2.error}"))
    raw_data2 = fetch_result2.or_raise()

    combined_data = raw_data1 + raw_data2

    # Step 2: Process combined data (cached)
    process_result = await process_data_heavy(combined_data)
    if not process_result.ok:
        print(f"[Orchestrator] Failed to process data: {process_result.error}")
        return Result.Error(Exception(f"Pipeline failed at process_data_heavy: {process_result.error}"))
    processed_output = process_result.or_raise()

    # Step 3: Send notification (idempotent, specific queue)
    recipient = config.get("notification_email", "admin@example.com")
    notification_subject = "Data Pipeline Complete"
    notification_body = f"Your data pipeline finished successfully. Output: {processed_output[:50]}..."
    
    notification_result = await send_notification(recipient, notification_subject, notification_body)
    if not notification_result.ok:
        print(f"[Orchestrator] Failed to send notification: {notification_result.error}")
        # Optionally, fail the whole pipeline or just log the notification error
        return Result.Error(Exception(f"Pipeline completed with notification error: {notification_result.error}"))

    print("[Orchestrator] Complex data pipeline finished successfully")
    return Result.Ok({"final_output": processed_output, "notification_sent": notification_result.or_raise()})

# --- Main Execution Logic ---

async def run_pipeline(executor: Stent, config: dict, expiry_str: str | None = None):
    print(f"\n--- Running Pipeline with config: {config} ---")
    execution_id = await executor.dispatch(complex_data_pipeline, config, expiry=expiry_str)
    print(f"Dispatched pipeline execution ID: {execution_id}")

    while True:
        state = await executor.state_of(execution_id)
        print(f"  State: {state.state}, Progress: {state.progress_str}")
        if state.state in ("completed", "failed", "timed_out"):
            break
        await asyncio.sleep(0.5)
    
    try:
        final_result = await executor.result_of(execution_id)
        print(f"  Final Result: {final_result}")
    except Exception as e:
        print(f"  Error retrieving result: {e}")
    print("---------------------------------------------------")
    return state.state

async def main():
    db_path = "complex_pipeline.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    
    Stent.mute_async_sleep_notifications = True
    executor = Stent(
        backend=backend, 
        notification_backend=Stent.notifications.RedisBackend("redis://localhost:6379")
        )
    
    # Start a worker that listens to all queues for now
    print("Starting a general worker...")
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.2))

    # Scenario 1: Successful run
    await run_pipeline(executor, {"url1": "http://good.com/dataA", "notification_email": "success@example.com"})
    
    # Scenario 2: Simulate failure in fetch_external_data (retryable)
    # The worker will retry 'fetch_external_data' for 'fail.com' twice (total 3 attempts).
    # After 3 failures, it will be marked failed, and orchestrator will catch it.
    await run_pipeline(executor, {"url1": "http://fail.com/data", "notification_email": "success@example.com"})

    # Scenario 3: Cached processing - should be faster
    # First run: process data normally
    await run_pipeline(executor, {"url1": "http://cache.com/data1", "url2": "http://cache.com/data2", "notification_email": "cache@example.com"})
    # Second run with same data for processing (process_data_heavy is cached=True) should hit cache
    await run_pipeline(executor, {"url1": "http://cache.com/data1", "url2": "http://cache.com/data2", "notification_email": "cache_again@example.com"})
    
    # Scenario 4: expiry an execution
    await run_pipeline(executor, {"url1": "http://slow.com/data", "url2": "http://slow.com/data"}, expiry_str="0.1s") # Will expiry due to long process_data_heavy

    # Scenario 5: Notification failure (non-retryable within send_notification if it expects clean recipient)
    # Note: send_notification only has basic exception handling, not specific retry_for for ValueError
    # But if the orchestrator is set to retry this, it will.
    # For this example, send_notification is idempotent, but if it fails, it will be marked failed.
    await run_pipeline(executor, {"notification_email": "error_recipient@example.com"})

    print("Shutting down worker.")
    worker_task.cancel()
    try:
        await worker_task
    except asyncio.CancelledError:
        pass
    print("Examples finished.")

if __name__ == "__main__":
    asyncio.run(main())



import hashlib
import logging
from typing import Any # Import Any
from stent.utils.serialization import Serializer, JsonSerializer

logger = logging.getLogger(__name__)

def default_idempotency_key(
    fn_name: str,
    version: str | None,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    *,
    serializer: Serializer | None = None,
) -> str:
    if serializer is None:
        serializer = JsonSerializer()
    
    # Simple structure to hash
    data = {
        "fn": fn_name,
        "version": version,
        "args": args,
        "kwargs": kwargs,
    }
    
    try:
        payload = serializer.dumps(data)
    except Exception:
        # Fallback if args aren't serializable by the default serializer
        payload = str(data).encode("utf-8")

    key = hashlib.sha256(payload).hexdigest()
    logger.debug(f"Generated idempotency key for {fn_name}: {key}")
    return key


from datetime import datetime, timedelta, timezone
import re

def parse_duration(duration: str | dict | timedelta) -> timedelta:
    if isinstance(duration, timedelta):
        return duration
        
    if isinstance(duration, dict):
        return timedelta(**duration)
        
    if not isinstance(duration, str):
        raise TypeError(f"Duration must be str, dict, or timedelta, got {type(duration)}")

    text = duration.strip()
    if not text:
        raise ValueError(f"Invalid duration string: {duration}")

    # Parse composite duration strings like "2d8h", "1h30m", "10s"
    token_re = re.compile(r"(\d+(?:\.\d*)?)([smhdw])")
    total_seconds = 0.0
    pos = 0
    for match in token_re.finditer(text):
        if match.start() != pos:
            raise ValueError(f"Invalid duration string: {duration}")

        value_str, unit = match.groups()
        value = float(value_str)
        if unit == 's':
            total_seconds += value
        elif unit == 'm':
            total_seconds += value * 60
        elif unit == 'h':
            total_seconds += value * 3600
        elif unit == 'd':
            total_seconds += value * 86400
        elif unit == 'w':
            total_seconds += value * 604800

        pos = match.end()

    if pos == 0 or pos != len(text):
        raise ValueError(f"Invalid duration string: {duration}")

    return timedelta(seconds=total_seconds)

def now_utc() -> datetime:
    return datetime.now(timezone.utc)



from typing import Any
import asyncio
import random
import logging
from stent import Stent

# Configure logging to see what's happening
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# 1. Define Activities
@Stent.durable()
async def fetch_url(url: str) -> str:
    """Simulates fetching a webpage with variable latency."""
    # Simulate network latency (0.5 to 2.0 seconds)
    # We use Stent.sleep to be "good citizens" and release the worker
    delay = random.uniform(0.5, 2.0)
    await Stent.sleep(f"{delay:.2f}s")
    
    if "error" in url:
        raise ValueError(f"404 Not Found: {url}")
        
    return f"<html><body><h1>Content of {url}</h1><p>Some interesting text here.</p></body></html>"

@Stent.durable()
async def analyze_sentiment(content: str) -> dict:
    """Simulates CPU-intensive analysis."""
    print(f"[Analysis] Analyzing sentiment: {content}...")
    await asyncio.sleep(0.1) # Simulate CPU work
    return {
        "score": random.randint(0, 100),
        "keywords": ["stent", "python", "durable"]
    }

# 2. Define Sub-Orchestrator
@Stent.durable()
async def process_page(url: str) -> dict:
    """
    Orchestrates the processing of a single page.
    This is a sub-workflow that will be run in parallel for multiple URLs.
    """
    print(f"[Page] Processing {url}...")
    
    # Call activity 1
    content = await fetch_url(url)
    
    # Call activity 2
    analysis = await analyze_sentiment(content)
    
    return {
        "url": url,
        "content_length": len(content),
        "analysis": analysis
    }

# 3. Define Main Orchestrator
@Stent.durable()
async def site_crawler(urls: list[str]) -> dict:
    """
    Main workflow that fans out to process multiple pages and fans in to aggregate results.
    """
    print(f"[Crawler] Starting crawl for {len(urls)} URLs")
    
    # Fan-out: Dispatch a sub-orchestrator for each URL.
    # We create the tasks but await them all at once using gather.
    tasks = [process_page(url) for url in urls]
    
    # Fan-in: Wait for all to complete.
    # return_exceptions=True allows the workflow to continue even if some pages fail.
    results: list[Any | BaseException] = await Stent.gather(*tasks, return_exceptions=True)
    
    # Aggregation Logic
    report: dict[str, Any] = {
        "total_urls": len(urls),
        "successful": 0,
        "failed": 0,
        "average_score": 0.0,
        "failures": [],
        "processed_data": []
    }
    
    total_score = 0
    
    for url, res in zip(urls, results):
        if isinstance(res, Exception) or isinstance(res, BaseException):
            report["failed"] += 1
            report["failures"].append({"url": url, "error": str(res)})
        else:
            report["successful"] += 1
            report["processed_data"].append(res)
            total_score += res["analysis"]["score"]
            
    if report["successful"] > 0:
        report["average_score"] = total_score / report["successful"]
        
    return report

# 4. Run the example
async def main():
    # Setup backend and executor
    backend = Stent.backends.SQLiteBackend("complex_gather.sqlite")
    await backend.init_db()
    executor = Stent(backend=backend)
    
    # Start worker in background (high concurrency for parallel tasks)
    worker_task = asyncio.create_task(executor.serve(max_concurrency=20))
    
    try:
        urls = [
            "https://example.com/page1",
            "https://example.com/page2",
            "https://example.com/error_page", # This will fail
            "https://example.com/page3",
            "https://example.com/page4",
        ]
        
        print("Dispatching workflow...")
        exec_id = await executor.dispatch(site_crawler, urls)
        print(f"Workflow dispatched: {exec_id}")
        
        # Poll for completion
        result = await executor.wait_for(exec_id, expiry=30.0)
        
        if result.ok:
            import json
            print("\nWorkflow Completed Successfully!")
            print(json.dumps(result.value, indent=2))
        else:
            print(f"\nWorkflow Failed: {result.error}")
            
    finally:
        worker_task.cancel()
        try:
            await worker_task
        except asyncio.CancelledError:
            pass
        # Cleanup
        import os
        if os.path.exists("complex_gather.sqlite"):
            os.remove("complex_gather.sqlite")

if __name__ == "__main__":
    asyncio.run(main())



from typing import Protocol, Any, Dict
import json
import pickle
import base64
import traceback
from stent.core import Result, RetryPolicy


# Registry of known exception classes for deserialization
# Maps fully qualified class name -> class
_EXCEPTION_REGISTRY: Dict[str, type[BaseException]] = {
    # Built-in exceptions
    "Exception": Exception,
    "ValueError": ValueError,
    "TypeError": TypeError,
    "KeyError": KeyError,
    "IndexError": IndexError,
    "AttributeError": AttributeError,
    "RuntimeError": RuntimeError,
    "IOError": IOError,
    "OSError": OSError,
    "FileNotFoundError": FileNotFoundError,
    "PermissionError": PermissionError,
    "TimeoutError": TimeoutError,
    "ConnectionError": ConnectionError,
    "NotImplementedError": NotImplementedError,
    "StopIteration": StopIteration,
    "AssertionError": AssertionError,
    "ImportError": ImportError,
    "ModuleNotFoundError": ModuleNotFoundError,
    "ZeroDivisionError": ZeroDivisionError,
    "OverflowError": OverflowError,
    "MemoryError": MemoryError,
    "RecursionError": RecursionError,
    "SystemError": SystemError,
    "UnicodeError": UnicodeError,
    "UnicodeDecodeError": UnicodeDecodeError,
    "UnicodeEncodeError": UnicodeEncodeError,
}


def register_exception(cls: type[BaseException], name: str | None = None) -> None:
    """Register an exception class for JSON deserialization.
    
    Args:
        cls: The exception class to register.
        name: Optional name to register under. Defaults to class __name__.
    """
    key = name or cls.__name__
    _EXCEPTION_REGISTRY[key] = cls


def get_exception_class(name: str) -> type[BaseException]:
    """Get an exception class from the registry, falling back to Exception."""
    return _EXCEPTION_REGISTRY.get(name, Exception)


def _extract_exception_attributes(exc: BaseException) -> Dict[str, Any]:
    """Extract custom attributes from an exception.
    
    Returns a dict of attributes that are not standard exception internals.
    """
    # Standard exception attributes to skip
    skip_attrs = {
        "args", "__dict__", "__cause__", "__context__", "__suppress_context__",
        "__traceback__", "__notes__", "add_note", "with_traceback",
    }
    
    attrs: Dict[str, Any] = {}
    for key in dir(exc):
        if key.startswith("_") or key in skip_attrs:
            continue
        try:
            val = getattr(exc, key)
            # Skip methods/callables
            if callable(val):
                continue
            # Try to ensure it's JSON-serializable
            json.dumps(val)
            attrs[key] = val
        except (TypeError, ValueError, AttributeError):
            # Not serializable, skip
            pass
    
    return attrs


def _format_traceback(exc: BaseException) -> str | None:
    """Format exception traceback as a string."""
    if exc.__traceback__ is None:
        return None
    return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))


class Serializer(Protocol):
    def dumps(self, obj: Any) -> bytes: ...
    def loads(self, data: bytes) -> Any: ...


class PickleSerializer(Serializer):
    def dumps(self, obj: Any) -> bytes:
        return pickle.dumps(obj)

    def loads(self, data: bytes) -> Any:
        return pickle.loads(data)


class JsonSerializer(Serializer):
    def dumps(self, obj: Any) -> bytes:
        return json.dumps(obj, default=self._default).encode("utf-8")

    def loads(self, data: bytes) -> Any:
        return json.loads(data.decode("utf-8"), object_hook=self._object_hook)

    def _default(self, obj: Any) -> Any:
        if isinstance(obj, Result):
            return {
                "__type__": "Result",
                "ok": obj.ok,
                "value": obj.value,
                "error": obj.error
            }
        if isinstance(obj, RetryPolicy):
            return {
                "__type__": "RetryPolicy",
                "max_attempts": obj.max_attempts,
                "backoff_factor": obj.backoff_factor,
                "initial_delay": obj.initial_delay,
                "max_delay": obj.max_delay,
                "jitter": obj.jitter,
                "retry_for": [exc.__name__ for exc in obj.retry_for],
            }
        if isinstance(obj, BaseException):
            return {
                "__type__": "Exception",
                "cls": obj.__class__.__name__,
                "message": str(obj),
                "traceback": _format_traceback(obj),
                "attributes": _extract_exception_attributes(obj),
            }
        # bytes handling for args/kwargs if they were pre-serialized
        if isinstance(obj, bytes):
            return {"__type__": "bytes", "data": base64.b64encode(obj).decode("ascii")}
        return str(obj)

    def _object_hook(self, dct: Dict[str, Any]) -> Any:
        if "__type__" in dct:
            t = dct["__type__"]
            if t == "Result":
                return Result(ok=dct["ok"], value=dct["value"], error=dct["error"])
            if t == "Exception":
                return _deserialize_exception(dct)
            if t == "RetryPolicy":
                return _deserialize_retry_policy(dct)
            if t == "bytes":
                return base64.b64decode(dct["data"].encode("ascii"))
        return dct


class DeserializedException(Exception):
    """Wrapper for deserialized exceptions that preserves original info."""
    
    def __init__(
        self,
        original_cls: str,
        message: str,
        original_traceback: str | None = None,
        attributes: Dict[str, Any] | None = None,
    ):
        super().__init__(message)
        self.original_cls = original_cls
        self.original_traceback = original_traceback
        self.original_attributes = attributes or {}
        
        # Copy attributes onto the exception
        for key, val in self.original_attributes.items():
            if not hasattr(self, key):
                setattr(self, key, val)
    
    def __str__(self) -> str:
        return f"{self.original_cls}: {super().__str__()}"
    
    def __repr__(self) -> str:
        return f"DeserializedException({self.original_cls!r}, {super().__str__()!r})"
    
    def format_original_traceback(self) -> str:
        """Return the original traceback if available, or a placeholder."""
        if self.original_traceback:
            return self.original_traceback
        return f"(no traceback available for {self.original_cls})"


def _deserialize_exception(dct: Dict[str, Any]) -> BaseException:
    """Deserialize an exception from its JSON representation.
    
    Attempts to reconstruct the original exception type if registered,
    otherwise returns a DeserializedException wrapper.
    """
    cls_name = dct.get("cls", "Exception")
    message = dct.get("message", "")
    tb_str = dct.get("traceback")
    attributes = dct.get("attributes", {})
    
    exc_class = _EXCEPTION_REGISTRY.get(cls_name)
    
    if exc_class is not None and exc_class is not Exception:
        # Try to instantiate the actual exception class
        try:
            exc = exc_class(message)
            # Attach custom attributes
            for key, val in attributes.items():
                try:
                    setattr(exc, key, val)
                except AttributeError:
                    pass
            # Store traceback string as custom attribute
            if tb_str:
                exc._original_traceback = tb_str  # type: ignore
            return exc
        except Exception:
            # Fall through to DeserializedException
            pass
    
    # Return wrapped exception preserving all info
    return DeserializedException(
        original_cls=cls_name,
        message=message,
        original_traceback=tb_str,
        attributes=attributes,
    )


def _deserialize_retry_policy(dct: Dict[str, Any]) -> RetryPolicy:
    """Deserialize a RetryPolicy from its JSON representation."""
    retry_for_names = dct.get("retry_for", ["Exception"])
    
    # Look up exception classes from registry, fall back to Exception
    retry_for_classes: list[type[BaseException]] = []
    for name in retry_for_names:
        exc_class = _EXCEPTION_REGISTRY.get(name, Exception)
        if exc_class not in retry_for_classes:
            retry_for_classes.append(exc_class)
    
    # Ensure we have at least one class
    if not retry_for_classes:
        retry_for_classes = [Exception]
    
    return RetryPolicy(
        max_attempts=dct.get("max_attempts", 3),
        backoff_factor=dct.get("backoff_factor", 2.0),
        initial_delay=dct.get("initial_delay", 1.0),
        max_delay=dct.get("max_delay", 60.0),
        jitter=dct.get("jitter", 0.1),
        retry_for=tuple(retry_for_classes),
    )
